#!/usr/bin/env lua
local coaf=require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"

coaf.ChDir(coaf.GetAppDir(arg))

local socket = require "cosocket.core"
socket.default_close_on_exec = true

local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
local service = require "coaf.service"
local event = require "coaf.event"
local cjson = require "cjson"
local httpclient = require "httpclient"
local avahi = require "coserver.discovery_tools.avahi"

---- 日志模块
local withLog = false
for _x=1, #arg do
	if arg[_x] == "--log" then
		withLog = true
		break
	end
end

if withLog then
	-- ** initialize the log file **
	local roll = require "coaf.log.rollfile"
	os.execute( "/bin/mkdir -p /var/log/" )
	local logF = roll.Create("/var/log/DiscoveryRE.log")
	if not logF or not logF:init() then
		_dbg.Error("TALK_SERVICE", "Fail to open log file: /var/log/DiscoveryRE.log, so the log messages should be output to console.")
		withLog = false
	else
		_dbg.AddLogger("logfile", logF, nil, nil)
	end
	-- ** log **
end

coaf.Init( "DiscoveryRE" )

if withLog then
	local stdio = require "coaf.log.stdio"
	stdio.SetDefaultFilterName( "DiscoveryRE" )
	stdio.Start()
end

service.InitSocket()
client.InitSocket( socket, adapter )

local function dict_to_list(dict)
    local list = {}
    for _, v in pairs(dict) do
        table.insert(list, v)
    end
    return list
end

adapter.timeoutUnit = 1000
local function GetProxy(objPath, service, loc, timeout, socket_addr)
	local proxy
	local cli

	cli = client.New( service, loc, socket_addr, timeout or 5 ) --5 seconds timeout.
	if not cli then
		_dbg.Error( "CO", "Fail to create connection to service '" .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		_dbg.Error( "CO", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

local mod = {}
local forbidden = {}
local DiscoverDict = {}
function mod.GetList()
    -- local proxy = GetProxy("/discovery", "_CO_DAEMON.coserver")
    -- if proxy then
    --     local is_ok, list = proxy:GetDiscoveryList()
    --     proxy:Destroy()
    --     if is_ok then
    --         local existDict = {}
    --         for _, v in ipairs(list) do
    --             if v.status ~= "trying" then
    --                 existDict[v.address] = true
    --                 DiscoverDict[v.address] = DiscoverDict[v.address] or {
    --                     address = v.address,
    --                     name = v.options and v.options.name,
    --                 }
    --             end
    --         end
    --         for k, v in pairs(DiscoverDict) do
    --             if not existDict[v.address] then
    --                 DiscoverDict[k] = nil
    --                 forbidden[k] = nil
    --             end
    --         end
    --     end
    -- end
    return dict_to_list(DiscoverDict)
end

local function getHTTPResponse(path)
    local r, body = httpclient.post(path, "", 2000, 147250)
    if r == 0 then
        while true do
            coaf.Sleep(0.1)
            r, body = httpclient.result( 147250 )
            if r ~= -1 then
                break
            end
        end
    end
    return r, body
end

local function GetAMBASerialNo(v)
    local status, body = getHTTPResponse("http://" .. v.address .. "/api/v1/deviceInfo.lua")
    if status == 0 and body ~= "" then
        local skip = pcall(function()
            body = cjson.decode(body)
        end)
        if not skip then
            _dbg.Error("DISCOVERY", "Decode json data failed(Amba). Address:" .. "http://" .. v.address .. "/api/v1/deviceInfo.lua")
            return
        end
        if not body.Result or body.Result ~= 200 then
            _dbg.Error("DISCOVERY", "Not Amba device return")
            return
        end

        v.serial = body.Data and body.Data.SERIAL_NO
        v.name = body.Data and body.Data.HOST_NAME
        v.oemType = body.Data and body.Data.DEVICE_OEM_TYPE or ""
        v.deviceType = "amba"
        return true
    end
end

local function GetHISISerialNo(v)
    local status, body = getHTTPResponse("http://" .. v.address .. "/api/info/get.json")
    if status == 0 and body ~= "" then
        local skip = pcall(function()
            body = cjson.decode(body)
        end)
        if not skip then
            _dbg.Error("DISCOVERY", "Decode json data failed(HiSi). Address:" .. "http://" .. v.address .. "/api/info/get.json")
            return
        end
        if not body.result or body.result ~= "ok" then
            _dbg.Error("DISCOVERY", "Not HiSi device return")
            return
        end
        v.serial = body.data and body.data.serialNumber
        v.name = body.data and body.data.name
        v.deviceType = "hisi"
        return true
    end
end

local function GetZYNQSerialNo(v)
    local status, body = getHTTPResponse("http://" .. v.address .. "/api/device/status.json")
    if status == 0 and body ~= "" then
        local skip = pcall(function()
            body = cjson.decode(body)
        end)
        if not skip then
            _dbg.Error("DISCOVERY", "Decode json data failed(Zynq). Address:" .. "http://" .. v.address .. "/api/device/status.json")
            return
        end
        if not body.result or body.result ~= "ok" then
            _dbg.Error("DISCOVERY", "Not Zynq device return")
            return
        end
        v.serial = body.data and body.data.serial_number
        v.deviceType = "zynq"
        return true
    end
end

local function GetZYNQName(v)
    local status, body = getHTTPResponse("http://" .. v.address .. "/api/device/get.json")
    if status == 0 and body ~= "" then
        local skip = pcall(function()
            body = cjson.decode(body)
        end)
        if not skip then
            _dbg.Error("DISCOVERY", "Decode json data failed(Zynq). Address:" .. "http://" .. v.address .. "/api/device/get.json")
            return
        end
        if not body.result or body.result ~= "ok" then
            _dbg.Error("DISCOVERY", "Not Zynq device return 2")
            return
        end
        v.name = body.data and body.data.device_name
        return true
    end
end

local function GetDeviceName()
    local filename = "/etc/hostname"
    local f = io.open( filename, "r" )
	if f then
		local r = f:read("*l")
		f:close()
        if r then
		    return r
        end
    end
    return "KILOVIEW-DiscoveryRE"
end

local function GetDeviceMode()
    if not mod.DeviceMode or mod.DeviceMode == "" then
        local _done, features = pcall(require, "oem")
        if _done then
            if type(features) ~= "table" then
                mod.DeviceMode = _G.OEM_TYPE
            else
                mod.DeviceMode = features.OEM_TYPE
            end
        end
        if not mod.DeviceMode or mod.DeviceMode == "" then
            _done, features = pcall(require, "features")
            if _done then
                if type(features) ~= "table" then
                    mod.DeviceMode = _G.PRODUCT_ID
                else
                    mod.DeviceMode = features.PRODUCT_ID
                end
            end
        end
    end
    return mod.DeviceMode
end

local pl = require "pl.pretty"
local ip_families = {"IPV4"}
local function disc_callback(service_type, action, service_item)
    --_dbg.Info( "TALKBACK", "--disc_callback--:", service_type, action)
    --pl.dump(service_item)
    if action == "cleanup" then--TODO: Should me clear all services?
		return
    end

	if not service_item or not service_item.name or not service_item.protocol or not service_item.interface then
		_dbg.Error( "TALK.Discovery", "Invalid service content for event:", action, "[", service_type, "]" )
		return
    end
    --Ignore local service.
	if service_item.is_local then
		return
    end

    local p = service_item.protocol:upper()
    local i, fnd
    for i=1, #ip_families do
        if p == ip_families[i] then
            fnd = true
            break
        end
    end
    if not fnd then
        print("OCLLOG, not fnd and return")
        return
    end

	if action == "new" or action == "update" or action == "resolve" then
		if not service_item.resolve_info then -- or service_item.resolve_info.untrust then
			return
		end

        local is_ip = service_item.resolve_info.address:match("%d+%.%d+%.%d+%.%d+")
        if not is_ip then
            return
        end

		DiscoverDict[ service_item.resolve_info.address ] = DiscoverDict[ service_item.resolve_info.address ] or {
            address = service_item.resolve_info.address,
            name = service_item.resolve_info.hostname:gsub("%.local", "")
        }
		
	elseif action == "remove" then --or action == "resolve-failed" then
        print("OCLLOG remove the address = ", service_item.resolve_info.address)
		DiscoverDict[service_item.resolve_info.address] = nil
	end
end

function mod.Discovery()
    avahi.PublishService( "11223", GetDeviceName(), "_discovery._tcp", "12312", {group = "public"} )
    avahi.SetDiscoveryCallback( disc_callback )
    avahi.StartDiscovery("_discovery._tcp")
    return true
end


local isChecking
local checkTimer = thread.NewTimer(function()
    if isChecking then
        return
    end
    isChecking = true
    for _, v in pairs(DiscoverDict) do
        -- if (forbidden[v.address] or 0) <= 5 then
            if GetAMBASerialNo(v) or (v.deviceType == "amba")  then
                v.weburl = "/device/" .. v.address .. "/index.lp"
            elseif GetHISISerialNo(v) or (v.deviceType == "hisi") then
                v.weburl = "/device/" .. v.address .. "/dashbord"
            else
                local failed = GetZYNQSerialNo(v)
                local failed2 = GetZYNQName(v)
                if not failed or not failed2 then
                    forbidden[v.address] = (forbidden[v.address] or 0) + 1
                end
                v.weburl = ""
            end
        -- end
    end
    isChecking = nil
end, 3, 15)

local CreateTimer = thread.NewTimer(mod.Discovery, 10) --针对于海思系列，太快启动会导致 PublishService失败，PublishService返回值一直为true，不好判断。
local UpdateTimer = thread.NewTimer(function()
    avahi.StopDiscovery("_discovery._tcp")
    coaf.Sleep(2)
    avahi.PublishService( "11223", GetDeviceName(), "_discovery._tcp", "12312", {group = "public"} )
    avahi.SetDiscoveryCallback( disc_callback )
    avahi.StartDiscovery("_discovery._tcp")
end, 10)

local cosvc_Talk
local function Main(args)
	cosvc_Talk = service.New( "DiscoveryRE", "unix:/var/run/discoveryRE.sock" )
	if cosvc_Talk then
		_dbg.Info( "DiscoveryRE", "CoService of 'talk' is created." )
		cosvc_Talk:CreateCoServerProxy()
	else
		_dbg.Error( "DiscoveryRE", "Fail created CoService!" )
	end

    mod.GetList()
	cosvc_Talk:ExposeObject("/discovery", mod )
    if checkTimer then
        checkTimer:start()
    end
    if CreateTimer then
        CreateTimer:start()
    end
    if UpdateTimer then
        UpdateTimer:start()
    end
end

coaf.Start(Main, arg)

if checkTimer then
    checkTimer:stop()
end

if UpdateTimer then
    UpdateTimer:stop()
end

if CreateTimer then
    CreateTimer:stop()
end

event.free( "discoveryRE" )
_dbg.Print( "discoveryRE", "Service terminated." )
