-- KILOVIEW G1 products Audio/Video front end.
local GV7601 = require("chips.gv7601")
local FPGA = require("chips.overlay_fpga")
local periphery = require( "periphery" )
local hwChk = require( "kiloview.G1.hardwareCheck" )
local I2C = periphery.I2C

require "features"

----------------------------------------------------------------
local _M = {
	deviceType = "G1",
	deviceName = "HD/3G-SDI H.264 ENCODER",
	vendor = "KILOVIEW Electronics Co., Ltd.",
	chip = nil,
	currentAudioSrc = "DIGITAL",
	audSelIO = nil,
	audSelI2C = nil,
	sdFormat_10bit = true,
	lastVideoFormatVersion = 0,
	lastVideoFormat = {},
	lastAudioFormatVersion = 0,
	lastAudioFormat = {},
	analogAudioFormat = {
		_valid = true,
		audio_src = "LINE",
		sampling = 48000, --TODO: How can dynamic support?
		audio_signal_type = 3, --AKA analog
		channels = 2,
		bit_width = 16
	},
}
----------------------------------------------------------------
local function AlsaSelect( src )
	if src == "MIC" then
		os.execute( [[/usr/bin/amixer cset name='Capture Source' 1]] )
		--os.execute( [[/usr/bin/amixer cset name='Line Input Switch' off]] )
		--os.execute( [[/usr/bin/amixer cset name='Mic Input Switch' on]] )
	else
		os.execute( [[/usr/bin/amixer cset name='Capture Source' 0]] )
		--os.execute( [[/usr/bin/amixer cset name='Line Input Switch' on]] )
		--os.execute( [[/usr/bin/amixer cset name='Mic Input Switch' off]] )
	end
end
----------------------------------------------------------------
local function CompareFormatInfo( a, b )
	if a == b then
		return false
	end
	
	local k,v
	for k,v in pairs(a) do
		if v ~= b[k] then
			return true
		end
	end
	
	for k,v in pairs(b) do
		if v ~= a[k] then
			return true
		end
	end
	
	return false
end
----------------------------------------------------------------
function _M:Init()
	--1st, Check the hardware version, if >=2.0, that uses a special audio driver.
	hwChk.Check()
	if _G.AUDSEL_BY_FPGA then
		self.audSelIO = nil
		local DEVICE = "/dev/i2c-3"
		if not _G.ENABLE_CONVERTER then
			DEVICE = "/dev/i2c-0"
		end
		local done,err = pcall( function() self.audSelI2C = I2C( DEVICE ) end )
		if not done then
			if type(err) == "table" then
				print( "ERROR: Could not open I2C device " .. DEVICE .. " for audio selection. ErrMsg: ", err.message or "---" )
			else
				print( "ERROR: Could not open I2C device " .. DEVICE .. " for audio selection. ErrMsg: ", err )
			end
			self.audSelI2C = nil
		end
	else
		self.audSelI2C = nil

		local tmpIO = periphery.GPIO( 59, "out" )
		tmpIO:write(false)
		periphery.sleep_ms(200)		--Wait 5ms for Calc CRC
		tmpIO:write(true)
		tmpIO:close()

		self.audSelIO = periphery.GPIO( 73, "out" )
		if not self.audSelIO then
			print( "WARNING: Could not open GPIO73 controller for audio digital/analog selection!" )
		else
			self.audSelIO:write( false )	-- 0: Digital(Default); 1: Analog
		end
	end

	if not _G.ENABLE_CONVERTER then
		self.chip = FPGA( "/dev/i2c-0", 0xe4)
	else
		self.chip = GV7601( { cs=80, clk=77, dout=79, din=78, io_10b20b = 76, io_sdhd=74 } )
	end
	if not self.chip then
		print( "ERROR: Initialize G1 AVFE: Could not create GV7601 controller!" )
		return false
	end
	self.chip:Reset()
	return true
end
----------------------------------------------------------------
function _M:StepCheck( enableSwitchNextSource )
	if self.chip then
		self.chip:Step( enableSwitchNextSource )
		if CompareFormatInfo( self.chip.videoFormat, self.lastVideoFormat ) then	--Different
			self.lastVideoFormatVersion = self.lastVideoFormatVersion + 1
			self.lastVideoFormat = {}
			local k,v
			for k,v in pairs(self.chip.videoFormat) do
				self.lastVideoFormat[k] = v
			end
		end
		
		if CompareFormatInfo( self.chip.audioFormat, self.lastAudioFormat) then -- Different audio
			self.lastAudioFormatVersion = self.lastAudioFormatVersion + 1
			self.lastAudioFormat = {}
			local k,v
			for k, v in pairs(self.chip.audioFormat) do
				self.lastAudioFormat[k] = v
			end
		end
	end
end
----------------------------------------------------------------
function _M:ReportDetectingStatus( srcType )
	if srcType == "video" then
		return true, "SDI"
	elseif srcType == "audio" then
		return true, self.currentAudioSrc
	end
end
----------------------------------------------------------------
function _M:ReportVideoFormat( curVersion )
	--if not curVersion or curVersion ~= self.lastVideoFormatVersion then
		return self.lastVideoFormatVersion, self.lastVideoFormat
	--else
	--	return self.lastVideoFormatVersion, nil	--Don't return format
	--end
end
----------------------------------------------------------------
function _M:ReportAudioFormat( curVersion )
	if self.savedAudioSrc ~= self.currentAudioSrc then
		self.savedAudioSrc = self.currentAudioSrc
		self.lastAudioFormatVersion = self.lastAudioFormatVersion + 1
		self.analogAudioFormat.audio_src = self.currentAudioSrc
	end

	if self.currentAudioSrc == "DIGITAL" then
		return self.lastAudioFormatVersion, self.lastAudioFormat
	else
		return self.lastAudioFormatVersion, self.analogAudioFormat
	end
end
----------------------------------------------------------------
function _M:SelectVideoSource( src )
	return "SDI"
end
----------------------------------------------------------------
local function PatchI2SWordLen(len)
	if hwChk.hardwareVersionSwCode < 0x20 then
		local file = io.open( "/proc/k2e_audio/wlen", "w" )
		if file then
			file:write( "" .. len )
			file:close()
		end
	end
end

local function byte2bin(n)
	local t = {}
	for i=7,0,-1 do
		t[#t+1] = math.floor(n / 2^i)
		n = n % 2^i
	end
	return table.concat(t)
end

function _M:SelectAudioSource( src, audioChannel )
	--TODO: Select audio source (SDI digital/Analog line-in/Analog Mic-in)
	src = src:upper()
	if src == "DIGITAL" or src == "SDI" or src == "AUTO" or src == "DEFAULT" then
		self.currentAudioSrc = "DIGITAL"
		if self.audSelIO then
			self.audSelIO:write( true )		-- 1: Digital
		elseif self.audSelI2C then
			if not _G.ENABLE_CONVERTER then
				audioChannel = audioChannel or 1
				print("[DEBUG] Audio select: (SDI) reg[0xa7] = ", tonumber( byte2bin(audioChannel - 1) .. 11, 2))
				self.audSelI2C:transfer( 0xe4 >> 1, {{0xa7, tonumber( byte2bin(audioChannel -1) .. 11, 2)}})
			else
				print( "[DEBUG] Audio select: (SDI) reg[0x09] = 0x01" )
				self.audSelI2C:transfer( 0xf4 >> 1, {{0x09, 0x01}} )
			end
		end
		PatchI2SWordLen(32)

	elseif src == "ANALOG" or src == "LINE" then
		self.currentAudioSrc = "LINE"
		if self.audSelIO then
			self.audSelIO:write( false )		-- 0: Analog
		elseif self.audSelI2C then
			if not _G.ENABLE_CONVERTER then
				print("[DEBUG] Audio select: (LINE) reg[0xa7] = ", tonumber( "10", 2))
				self.audSelI2C:transfer( 0xe4 >> 1, {{0xa7, 0x02}})
			else
				print( "[DEBUG] Audio select: (LINE) reg[0x09] = 0x00" )
				self.audSelI2C:transfer( 0xf4 >> 1, {{0x09, 0x00}} )
			end
		end
		AlsaSelect( "LINE" )
		PatchI2SWordLen(64)

	elseif src == "MIC" then
		self.currentAudioSrc = "MIC"
		if self.audSelIO then
			self.audSelIO:write( false )		-- 0: Analog
		elseif self.audSelI2C then
			print( "[DEBUG] Audio select: (MIC) reg[0x09] = 0x00" )
			self.audSelI2C:transfer( 0xf4 >> 1, {{0x09, 0x00}} )
		end
		AlsaSelect( "MIC" )
		PatchI2SWordLen(64)

	elseif src == "AES" then
		self.currentAudioSrc = "AES"
		if self.audSelIO then
			self.audSelIO:write( false )		-- 0: Analog
		elseif self.audSelI2C then
			print( "[DEBUG] Audio select: (AES) reg[0x09] = 0x02" )
			self.audSelI2C:transfer( 0xf4 >> 1, {{0x09, 0x02}} )
		end
		--Does this required?
		--AlsaSelect( "LINE" )
		PatchI2SWordLen(32)
	end
	return self.currentAudioSrc
end
----------------------------------------------------------------
function _M:SetVideoParameter( pmName, pmValue, src )
	--Nothing
end
----------------------------------------------------------------
function _M:SetAudioParameter( pmName, pmValue, src )
	if not src then
		src = self.currentAudioSrc
	end
	
	if src == "DIGITAL" or src == "SDI" or src == "AES" then
		return --Nothing
	end
	
	if pmName == "gain" then	--Adjust analog audio gain
		if src == "ANALOG" or src == "LINE" then
			pmValue = tonumber(pmValue)	--As DB
			local dbSteps = math.floor(pmValue/0.75)	--For AIC23 chip, value 1 means 0.75db
			if dbSteps > 8 then
				dbSteps = 8
			elseif dbSteps < -23 then
				dbSteps = -23
			end
			
			local dbVal = 23 + dbSteps
			os.execute( [[/usr/bin/amixer cset name='Line Input Volume' ]] .. dbVal )
		elseif src == "MIC" then
			pmValue = tonumber(pmValue)
			if pmValue < 3 then	-- <3db, as default mic gain (0db)
				os.execute( [[/usr/bin/amixer cset name='Mic Booster Switch' off]] )
			else	-- >=3db, Mic boost (so with 20db gain)
				os.execute( [[/usr/bin/amixer cset name='Mic Booster Switch' on]] )
			end
		end
	elseif pmName == "line_bypass" then
		if pmValue == 0 then
			os.execute( [[/usr/bin/amixer cset name='Output Mixer Line Bypass Switch' off]] )
		else
			os.execute( [[/usr/bin/amixer cset name='Output Mixer Line Bypass Switch' on]] )
		end
	end
end
----------------------------------------------------------------
function _M:GetSupportVideoSources()
	return { ["SDI"] = "SD/HD/3G-SDI Video Input" }
end
----------------------------------------------------------------
function _M:GetSupportAudioSources()
	return { ["SDI"] = "SDI AES/EBU Digital Audio", ["LINE"] = "Analog Line Input" --[[, ["MIC"] = "Analog MIC Input"]] }
end
----------------------------------------------------------------
return _M

