-----------------------------------------------------------------------------------------------------
-- Overlay FPGA chip controller.
-- Copyright (C) 2016, KILOVIEW Elecontrics Co., Ltd.
--
-- Author: 	Jacob Zuo, 2018/5/4
-----------------------------------------------------------------------------------------------------

local periphery = require "periphery"
local I2C = periphery.I2C
local GPIO = periphery.GPIO
--For debug only.
--local pretty = require "pl.pretty"

local sdi_formats = require "sdi_formats"
sdi_formats.Init()

local fpga_format_map = {
	[0] = {
		[4] = 0x18, --576i
		[5] = 0x16, --480i
	},

	[1] = {
		[1] = 0x08, --720p24
		[2] = 0x06, --720p25
		[3] = 0x02, --720p30
		[6] = 0x04, --720p50
		[7] = 0x00, --720p60
	},

	[2] = {
		[4] = 0x15, --1035i50 (but only 60 support)
		[5] = 0x15, --1035i60
	},

	[3] = {
		[1] = 0x10, --1080p24
		[2] = 0x0d, --1080p25
		[3] = 0x0b, --1080p30
		[4] = 0x14, --1080i50
		[5] = 0x0a, --1080i60
		[6] = 0x2d, --1080p50
		[7] = 0x2b, --1080p60
	}
}

------------------------------------------------------
-- The factory function to create a OVFPGA handle object
------------------------------------------------------
local function CreateOVFPGA( i2c_dev, baseAddr )
	i2c_dev = i2c_dev or "/dev/i2c-0"
	
	local BASE_MAP = baseAddr or 0xe2

	print( "Creating OVFPGA control object at " .. string.format("%02X",BASE_MAP) .. "@" .. i2c_dev .. "..." )

	------------------------------------------------------
	-- Object
	------------------------------------------------------
	local OVFPGA = {
		currentSource = "SDI",
		currentSignal = false,
		freeRun = false,
		currentFormat = nil,
		forceReassignFormat = false,
		lastFormatRecord = {},
		videoFormat = { _valid = false },
		audioFormat = { _valid = false, audio_src = "SDI", audio_signal_type = 0 }, --type 0 aka. HDMI
		orgAudioFormat = { _valid = false },
		nostandardFormatInfo = {},
		i2c_handle = nil,
	}

	-------------------------------------------------------
	-- I2C and GPIO openning
	-------------------------------------------------------
	local done,err = pcall( function() OVFPGA.i2c_handle = I2C( i2c_dev ) end )
	if not done then
		if type(err) == "table" then
			print( "ERROR: Could not open I2C device for OVFPGA controlling. ErrMsg:", err.message or "---" )
		else
			print( "ERROR: Could not open I2C device for OVFPGA controlling. ErrMsg:", err )
		end
		return nil, err
	end
	
	-------------------------------------------------------
	-- Object method: Register read
	-------------------------------------------------------
	function OVFPGA:ReadReg( addr, itemArr )
		if not self.i2c_handle then
			return nil, "Invalid I2C handle"
		end
		if not itemArr or type(itemArr) ~= "table" then
			itemArr = { 0xff, flags = I2C.I2C_M_RD }
		else
			if #itemArr < 1 then
				itemArr[1] = 0xff
			end
			if not itemArr.flags then
				itemArr.flags = I2C.I2C_M_RD
			else
				itemArr.flags = itemArr.flags | I2C.I2C_M_RD
			end
		end

		local msg = { {addr}, itemArr }
		local done,err = pcall( function() self.i2c_handle:transfer( BASE_MAP>>1, msg ) end )
		if not done then
			if type(err) == "table" then
				--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_MAP, addr), err.message or "---" )
				return nil, err.message or "---"
			else
				--print( string.format( "ReadReg fail: 0x%02x [0x%02x], Error:", BASE_MAP, addr), err )
				return nil, err
			end
		else
			return itemArr
		end
	end

	------------------------------------------------------
	-- Register write
	------------------------------------------------------
	function OVFPGA:WriteReg( addr, itemArr )
		if not self.i2c_handle then
			return false, "Invalid I2C handle"
		end
		if not itemArr or type(itemArr) ~= "table" or #itemArr < 1 then
			return false, "Invalid data"
		end

		table.insert( itemArr, 1, addr )
		local msg = { itemArr }
		local done,err = pcall( function() self.i2c_handle:transfer( BASE_MAP>>1, msg ) end )
		if not done then
			if type(err) == "table" then
				--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_MAP, addr), err.message or "---" )
				return nil, err.message or "---"
			else
				--print( string.format( "WriteReg fail: 0x%02x [0x%02x], Error:", BASE_MAP, addr), err )
				return nil, err
			end
		else
			return true
		end
	end

	------------------------------------------------------
	-- Soft reset the chip!
	------------------------------------------------------
	function OVFPGA:Reset()
		--Nothing
		self.currentSignal = false
		self.freeRun = false
		self.currentFormat = nil
		self.forceReassignFormat = false
		self.lastFormatRecord = {}		--Clear recorded format detection informations.
		self.videoFormat = {_valid = false}
		
		--TODO: Now fixed it.
		self.audioFormat = { 
			_valid = true,
			audio_src = "SDI",
			sampling = 48000, 
			audio_signal_type = 1, --AKA SDI
			channels = 2,
			bit_width = 16
		}
	end

	------------------------------------------------------
	-- Init the OVFPGA interrupts 
	------------------------------------------------------
	
	function OVFPGA:_InitInterrupts()
		--TODO: Nothing yet.
	end
	
	------------------------------------------------------
	-- Check SDI signal and type 
	------------------------------------------------------
	
	function OVFPGA:_CheckSignalAndType()
		local rslt = self:ReadReg( 0xa1, { 0 } )
		--[[
		local ver = self:ReadReg( 0xb8, {0} )
		if ver then
			ver = ver[1]
		end
		--]]
		local ver = 0
		if not rslt then
			self.currentSignal = false
			return false, ver, nil
		else
			if rslt[1] == 0 then
				return false, ver, nil
			else
				rslt = self:ReadReg( 0xa2, {0, 0} )
				if rslt then
					if rslt[1] == 0 then
						return true, ver, sdi_formats.VIRATE_SD
					elseif rslt[1] < 3 then
						return true, ver, sdi_formats.VIRATE_HD
					elseif rslt[2] >= 6 then
						return true, ver, sdi_formats.VIRATE_3G
					else
						return true, ver, sdi_formats.VIRATE_HD
					end
				else
					return false, ver, nil
				end
			end
		end
	end
	
	------------------------------------------------------
	
	function OVFPGA:ForceUpdate()
		self.lastFormatRecord = {}		--Clear the recorded format information.
	end

	------------------------------------------------------
	-- A step handler for OVFPGA processing.
	-- To fit the LUA working mode, you need call the 
	-- 'Step' periodically ( suggest per 200~500ms).
	-- This routine should do :
	--	* Signal status checking
	--	*	Auto format detecting
	--	*	Configue the chip to make it output BT.656/BT.601 signals
	--
	------------------------------------------------------
	
	function OVFPGA:Step()
		local oldSignal, signalVer, viRate
		oldSignal = self.currentSignal
		self.currentSignal, signalVer, viRate = self:_CheckSignalAndType()
		
		if oldSignal ~= self.currentSignal or signalVer ~= self.videoFormat.signal_version then
			print( "SDI INFO: SIGNAL=" .. (self.currentSignal and "<locked>" or "<unlocked>") .. ", Rate=" .. (viRate or "<unknown>") )
			print( "SDI signal loss version:", signalVer )
		end

		self.videoFormat.signal_version = signalVer
		if self.currentSignal then
			self.videoFormat.signal_locked = true	--Assume it be, but when detecting, it may changed
			self:_DetectCurrentFormat()
		else
			self.videoFormat.signal_locked = false
		end
	end


	------------------------------------------------------
	-- Detect current signal format
	------------------------------------------------------
	
	function OVFPGA:_DetectCurrentFormat(src)
		local std_id
		local rslt = self:ReadReg( 0xa2, {0, 0} )
		if not rslt then
			return false
		end

		local map1 = fpga_format_map[ rslt[1] ]
		if not map1 then
			self.videoFormat._valid = false
			return false
		end

		std_id = map1[rslt[2]]
		if not std_id then
			self.videoFormat._valid = false
			return false
		end

		if std_id ~= self.lastFormatRecord.std_id then
			print( "SDI INFO: Format STD-ID = " .. std_id )

			local formatInfo = sdi_formats.format_map[ std_id ]
			if formatInfo and #formatInfo >= 1 then
				formatInfo = formatInfo[1]
			end

			if formatInfo then
				self.lastFormatRecord.std_id = std_id
				print( "Matched format: ", formatInfo.formatId )
				self.videoFormat._valid = true
				self:_RecordVideoFormatInfo(src, formatInfo)
				return true
			else
				print( "*** Oh! No matched format for it, what's wrong?" )
			end
		else
			self.videoFormat._valid = true
			return true
		end
		
		self.videoFormat._valid = false
		--TODO: Other format checking but now not support
		return false
	end
	
	------------------------------------------------------
	-- According detected format information, record it to Amba based formation context.
	------------------------------------------------------
	
	function OVFPGA:_RecordVideoFormatInfo(src, formatInfo)
		if type(formatInfo) ~= "table" then
			return false
		end

		--pretty.dump( formatInfo )
		
		if self.videoFormat._formatInfo ~= formatInfo or (src and src ~= self.videoFormat.video_src) then
			self.videoFormat._lastFormatChanged = true
			self.videoFormat._formatInfo = formatInfo
			
			if src then
				self.videoFormat.video_src = src
			end
			
			--TODO:
			self.videoFormat.cap_start_x = 0
			self.videoFormat.cap_start_y = 0
			
			self.videoFormat.cap_cap_w = formatInfo.hAct
			self.videoFormat.line_width = formatInfo.hAct
			
			self.videoFormat.total_width = formatInfo.hTotal
			
			if formatInfo.vAct then
				self.videoFormat.cap_cap_h = formatInfo.vAct
				self.videoFormat.total_height = formatInfo.vTotal
				
				if formatInfo.interlaced then
					self.videoFormat.f0_height = formatInfo.vAct/2
					self.videoFormat.f1_height = formatInfo.vAct/2
				else
					self.videoFormat.f0_height = formatInfo.vAct
					self.videoFormat.f1_height = formatInfo.vAct
				end
			elseif formatInfo.vAct0 then
				self.videoFormat.f0_height = formatInfo.vAct0
				if formatInfo.interlaced then
					if formatInfo.vAct1 then
						self.videoFormat.f1_height = formatInfo.vAct1
						self.videoFormat.cap_cap_h = formatInfo.vAct0 + formatInfo.vAct1
						if self.videoFormat.cap_cap_h >= 486 and self.videoFormat.cap_cap_h <= 488 then
							self.videoFormat.cap_cap_h = 480	--This is a special format
						else
							self.videoFormat.cap_cap_h = self.videoFormat.cap_cap_h & 0xfffffffe
						end
						self.videoFormat.total_height = self.videoFormat.cap_cap_h	--??? total_height for interlace?
					else
						self.videoFormat.f1_height = formatInfo.vAct0
						self.videoFormat.cap_cap_h = formatInfo.vAct0 * 2
						self.videoFormat.total_height = formatInfo.vAct0 * 2
					end
				else
					self.videoFormat.f1_height = formatInfo.vAct0
					self.videoFormat.cap_cap_h = formatInfo.vAct0
					self.videoFormat.total_height = formatInfo.vTotal0
				end
			else
				self.videoFormat.f0_height = 0
				self.videoFormat.f1_height = 0
				self.videoFormat.cap_cap_h = 0	--This is wrong.
				self.videoFormat.total_height  = 0
			end
			
			--TODO:
			self.videoFormat.sync_start = 0
			
			if formatInfo.video_system then
				self.videoFormat.video_system = formatInfo.video_system
			else
				self.videoFormat.video_system = _G.AMBA_VIDEO_SYSTEM_AUTO or 0x0
			end
			
			if AMBA_VIDEO_FPS then
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS(formatInfo.frameRate)
			else
				self.videoFormat.frame_rate = AMBA_VIDEO_FPS_60 or formatInfo.frameRate
			end
			
			self.videoFormat.frame_rate_value = formatInfo.frameRate
			
			if formatInfo.aspect_ratio then
				self.videoFormat.aspect_ratio = formatInfo.aspect_ratio
			else
				self.videoFormat.aspect_ratio = _G.AMBA_VIDEO_RATIO_AUTO or 0
			end
			
			--TODO: Here, need fix this to support 8 bits/ BT.601 options.
			self.videoFormat.input_type = _G.AMBA_VIDEO_TYPE_YUV_656 or 2
			self.videoFormat.bit_resolution = 16
			if formatInfo.interlaced then
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_INTERLACE or 1
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_INTLC or 1
				self.videoFormat.interlaced = 1
			else
				self.videoFormat.video_format = _G.AMBA_VIDEO_FORMAT_PROGRESSIVE or 2
				self.videoFormat.input_format = _G.AMBA_VIN_INPUT_FORMAT_YUV_422_PROG or 2
				self.videoFormat.interlaced = 0
			end
			
			--TODO:
			self.videoFormat.hs_front_porch = 0
			self.videoFormat.hs_pulse_width = 0
			self.videoFormat.hs_back_porch = 0
			--TODO:
			self.videoFormat.f0_vs_front_porch = 0
			self.videoFormat.f0_vs_pulse_width = 0
			self.videoFormat.f0_vs_back_porch = 0
			--TODO:
			self.videoFormat.f1_vs_front_porch = 0
			self.videoFormat.f1_vs_pulse_width = 0
			self.videoFormat.f1_vs_back_porch = 0
		
			if formatInfo.tmdsFreq and formatInfo.tmdsFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.tmdsFreq 
			elseif formatInfo.pixelFreq and formatInfo.pixelFreq > 1000000 then
				self.videoFormat.pixel_freq = formatInfo.pixelFreq
			else
				self.videoFormat.pixel_freq = 0
			end
			
			self.videoFormat.format_code = formatInfo.formatId	-- Make the formatId as format-code.
			self.videoFormat.format_name = formatInfo.formatName or (self.videoFormat.cap_cap_w .. "x" .. self.videoFormat.cap_cap_h)
			self.videoFormat.video_signal_type = _G.VIDEO_SIGNAL_SDI or 2
			return true
		else
			return true	--No changed.
		end
	end

	------------------------------------------------------
	-- Detect SDI Audio
	------------------------------------------------------
	function OVFPGA:_DetectAudio(src)
		--TODO:
	end

	------------------------------------------------------
	-- Assume SDI customized format
	------------------------------------------------------
	function OVFPGA:_AssumeCustomFormat(src, BL, LCF, FCL, LCVS, interlaced)
		return false	--TODO
	end
	
	------------------------------------------------------
	return OVFPGA
end --END of CreateOVFPGA()

return CreateOVFPGA

