#!/usr/bin/env lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"

-- Change current dir to app location
coaf.ChDir(coaf.GetAppDir(arg))

local socket = require "cosocket.core"
local client = require "coaf.client"
local adapter = require "coaf.cli_adapter.simple"
local service = require "coaf.service"

local thread = require "coaf.thread"

local roll = require "coaf.log.rollfile"

local withLog = false
local _x
for _x=1, #arg do
	if arg[_x] == "--log" then
		withLog = true
		break
	end
end

if withLog then
	-- ** initialize the log file **
	local roll = require "coaf.log.rollfile"
	os.execute( "/bin/mkdir -p /data/log/" )
	local logF = roll.Create("/data/log/codec.log")
	if not logF or not logF:init() then
		_dbg.Error("CODEC", "Fail to open log file: /data/log/codec.log, so the log messages should be output to console.")
		withLog = false
	else
		_dbg.AddLogger("logfile", logF, nil, nil)
	end
	-- ** log **
end

coaf.Init( "codec_service" )

_G.print = function(...)
	_dbg.Info1( "CODEC", ... )
end

if withLog then
	local stdio = require "coaf.log.stdio"
	stdio.SetDefaultFilterName( "CODEC" )
	stdio.Start()
end

--------

--The global environments.
_G._ENV = {}

--------
print( ">>> Init: CoAF...")
service.InitSocket()
client.InitSocket( socket, adapter )
--------

local mo_codecs
local mo_encoders
local mo_encoder
local mo_osd
local mo_avio

local mo_blockloopout

local cosvc_Codec
local fast_timer
local period_timer

--------
--For OSD render.
local render_period = 0.2
local render_counter = 0
local function fastRenderCheck()
	mo_osd.PeriodRender()
	render_counter = render_counter + 1
	if render_counter * render_period >= 0.5 then
		mo_avio.Check()
		render_counter = 0
	end
end
--------
local function periodCheck()
	--mo_avio.Check()
	mo_codecs.Check()
	mo_encoders.Check()
	mo_encoder.Check()
	mo_encoder.NotifyStartEncoder("all")
end
--------
local function Main()
	print( "*** KILOVIEW AVIO/Codecs module running ***" )

	print( ">>> Loading: CODEC service...")
	mo_codecs = require "codecs"
	
	print( ">>> Loading: Encoder manager module...")
	mo_encoders = require "encoders"
	
	print( ">>> Loading: Encoder module...")
	mo_encoder = require "encoder"
	
	print( ">>> Loading: OSD module..." )
	mo_osd = require "osd"

	--------
	--@@@NOTE: !!!! Don't load AVIO before dbus service connection creation !!!!
	--Our AVIO module may need load MD3xx and that module should take more than 5 seconds
	--to init the chip. But, if 'systemd' found too long time to start the DBUS service, it send a 'SIGTERM'
	--signal to terminate this. Hmm, so bad!
	
	print( ">>> Loading: Video/Audio input controlling module...")
	mo_avio = require "avio"

	mo_blockloopout = require "block_loop_out"

	cosvc_Codec = service.New( "codec", "unix:/var/run/codec.sock" )
	if cosvc_Codec then
		_dbg.Info( "CODEC", "CoService of 'codec' is created." )
		cosvc_Codec:CreateCoServerProxy()
	else
		_dbg.Error( "CODEC", "Fail created CoService!" )
	end

	print( ">>> Loading: Modules initialization..." )
	mo_avio.Init( cosvc_Codec )
	mo_osd.Init( cosvc_Codec )
	mo_codecs.Init( cosvc_Codec )
	mo_encoders.Init( cosvc_Codec )
	mo_encoder.Init( cosvc_Codec )

	mo_blockloopout.Init(cosvc_Codec)

	mo_avio.Check()	--First check it.
	mo_encoder.NotifyStopEncoder("all")	--Stop all first, for first remained status reason.

	--Use thread-based timer instead of default coaf timer.
	--NOTE: thread-based timer requires CoAF version >= 1.01.
	fast_timer = thread.NewTimer( fastRenderCheck, 0.01, render_period)
	period_timer = thread.NewTimer( periodCheck, 0.01, 1 )

	if fast_timer then
		fast_timer:start()
	end
	if period_timer then
		period_timer:start()
	end
end

--------

coaf.Start(Main)

_dbg.Info( "CODEC", "Codec service is terminated." )

