require "kmCore"	--Media core. [ got KM_STREAMING global ]
----------------------------------------------------
local coaf = require "coaf.core"
local dbusCli = require "kiloview.dbusClients"
local getset = require "kiloview.getSetHelper"
local _DECL = require "kiloview.dbus.sunshine_record"

local clientSession
----------------------------------------------------
local _M = {
	allObjects = {},
	dbusClientSession = nil,
	recordProxy = nil,
	recordEventsProxy = nil,
	recorderRegistered = nil,
	recordGroup = nil
}

----------------------------------------------------
-- The signal events for recording service.
local function NotifyStartRecord( filesys, stream, seqId, format, path, fileprefix, limit_type, limit ) 
	local mod_streams = require "streams"

	print( "*** StartRecord:", filesys, stream, seqId, format, path, fileprefix, limit_type, limit )
	local streamIdx = 0
	local streamIdr = "00"
	if stream == "sub" or stream == "1" then
		streamIdx = 1
		streamIdr = "01"
	end

	if not _M.recorderRegistered then
		print( "***WARNING: The recorder is pending initialization, wait!" )
		_M.recordProxy:AckStartRecord( seqId, "started" )
		return
	end

	if _G.streamer.stream[streamIdx].enabled == 0 then
		print( "***WARNING: The stream of '" .. stream .. "' is disabled, so could not start recording!" )
		_M.recordProxy:AckStartRecord( seqId, "not-start" )
		return
	end

	local thisSvc = mod_streams.stream_objs[streamIdr]
	if not thisSvc then
		print( "***WARNING: The stream of '" .. stream .. "' is not exist, so could not start recording!" )
		_M.recordProxy:AckStartRecord( seqId, "not-exist" )
	end

	if not thisSvc.source.mediaSource then
		print( "*** ERROR: Invalid IAV source, so could not start the recording of stream '" .. stream .. "'" )
		_M.recordProxy:AckStartRecord( seqId, "no-source" )
		return
	end

	if thisSvc.recorder then --It's already started.
		_M.recordProxy:AckStartRecord( seqId, "started" )
		return
	end

	thisSvc.recorder = _M.createNew( filesys, format, streamIdx )
	if not thisSvc.recorder then
		print( "*** ERROR: Fail to create recorder for stream '" .. stream .. "', format=" .. format )
		_M.recordProxy:AckStartRecord( seqId, "start-fail" )
		return
	end

	if not thisSvc.recorder:BindInputMediaSource( thisSvc.source.mediaSource ) then
		print( "*** ERROR: Fail to bind media source for recording stream '" .. stream .. "', format=" .. format )
		_M.recordProxy:AckStartRecord( seqId, "start-fail" )
		return
	end

	thisSvc.recorder:SetRecordOptions( path, seqId, fileprefix, limit_type, limit )

	--NOTE here: Since 'StartRecord()' should do coaf proxy call, and it may suspend()/resume() current
	--           coroutine, this can't be done in this callback function, because the up caller is 
	--           C-DBus module, the C/C++ module can't resume the coroutine for Lua!
	coaf.ScheduleDelayedTask( function()
		if not thisSvc.recorder:StartRecord() then
			print( "*** ERROR: Fail to start recording for stream '" .. stream .. "', format=" .. format )
			_M.recordProxy:AckStartRecord( seqId, "start-fail" )
			return
		else
			_M.recordProxy:AckStartRecord( seqId, "started" )
		end
	end )
end       


local function NotifyStopRecord( stream, seqId )
	local mod_streams = require "streams"

	print( "*** StopRecord:", stream, seqId )

	local streamIdx = 0
	local streamIdr = "00"
	if stream == "sub" or stream == "1" then
		streamIdx = 1
		streamIdr = "01"
	end

	local thisSvc = mod_streams.stream_objs[streamIdr]
	if thisSvc and thisSvc.recorder then
		thisSvc.recorder:StopRecord()
		_M.free( thisSvc.recorder )
		thisSvc.recorder = nil
	end
	_M.recordProxy:AckStopRecord( seqId, "stopped" )
end

----------------------------------------------------
local function delayReconnectRecordDBus()                                                                              
	if not _M.recordProxy then                                                                                     
			_M.recordProxy = nil                                                                                   
			_M.recordProxy = _M.dbusClientSession:GetProxy( _DECL.RECORDM_OBJ, _DECL.RECORDM_BUSNAME, _DECL.iRecord )
	end                                                                                                            
	if _M.recordProxy then                                                                                         
			print( "***NICE: Success get recording DBUS proxy!" )                                                  
			return                                                                                                 
	end                                                                                                            
	print( "***ERROR: Fail to get recording DBUS proxy!" )                                                         
	coaf.NewTimer( delayRestartRecord, 10):start()--After 0.01 second, and per 50 seconds check.                   
end 

function _M.Init()
	local dbusCliSession = dbusCli.CreateClientSession()
	if not dbusCliSession then
		error( "Could not create client session for Encoder service (master using)" )
	end

	_M.dbusClientSession = dbusCliSession
	delayReconnectRecordDBus()

	_M.recordEventsProxy = dbusCliSession:GetProxy( _DECL.RECORDM_OBJ, _DECL.RECORDM_BUSNAME, _DECL.iRecordEvents )
	if not _M.recordEventsProxy then
		print( "*** WARNING: Could not get the recording proxies, so the recording function may be disabled!" )
	else
		_M.recordEventsProxy:ConnectSignal( "NotifyStartRecord", NotifyStartRecord )
		_M.recordEventsProxy:ConnectSignal( "NotifyStopRecord", NotifyStopRecord )
		--_M.recordProxy:RegisterWorker( "any" )
	end
	return true
end

----------------------------------------------------
function _M.createNew( fileSys, fileType, streamIdx )
	if not _M.dbusClientSession or not _M.recordProxy then
		print( "***ERROR: Un-initialized record module, invalid DBUS client session!" )
		return nil
	end

	local OBJ = {
		fileSys = fileSys or "vfat",
		fileType = fileType or "ts",
		streamIdx = streamIdx,
		filePrefix = "REC",
		recordSess = nil,
		mediaSource = nil,
		recordingStatus = "idle",
		location = nil,
		sequenceId = nil,
		limitionType = nil,
		limitValue = nil,
		lastError = nil,
		curSplitFile = nil,
		curSplitStatus = "idle",
		curFileSize = 0,
		curFileDur = 0
	}

	function OBJ:BindInputMediaSource( mediaSource )
		if self.recordingStatus == "recording" then
			print( "***ERROR: The media source is already bound and you must stop recording to bind again!" )
			return false
		end

		self.mediaSource = mediaSource
		return true
	end

	function OBJ:SetRecordOptions( location, sequenceId, fileprefix, limitionType, limitValue )
		if self.recordingStatus == "recording" then
			print( "***ERROR: The recording is on work, please stop it first to set recording options." )
			return false
		end

		self.filePrefix = fileprefix
		self.location = location
		self.sequenceId = sequenceId
		self.limitionType = limitionType or "none"
		self.limitValue = limitValue or "0"
		return true
	end

	function OBJ:StartRecord()
		if self.recordingStatus == "recording" or self.recordingStatus == "preparing" then
			return true
		end

		if not self.mediaSource then
			self.lastError = "invalid_media_source"
			print( "***ERROR: Invalid media source bound, could not start recording" )
			return false
		end

		if not self.location or not self.sequenceId then
			self.lastError = "invalid_options"
			print( "***ERROR: Invalid recording options!" )
			return false
		end

		if not _M.recordGroup then
			_M.recordGroup = KM_STREAMING.Recorder.createNew("recorder")
			if not _M.recordGroup then
				self.lastError = "fail_create_recorder"
				print( "***ERROR: Fail to create recorder group" )
				return false
			end
		end

		self.recordSess = _M.recordGroup:AddSession(self.sequenceId, self.mediaSource, 4*1024*1024, 64*1024)
		if not self.recordSess then
			print( "***ERROR: Fail to create a record session for sequence:", self.sequenceId)
			self.lastError = "fail_create_record_session"
			return false
		end

		local multiple = false
		local maxSize = "4G" --NOTE:TODO:This is a limitation for current device.
		if self.fileSys == "ntfs" or self.fileSys:match("ext") then
			maxSize = "0"
		end
		local maxTime = "0"
		if self.limitionType:match( "time" ) then
			if self.limitionType:match( "loop" ) then
				multiple = true
			elseif self.limitionType:match( "multiple" ) then
				multiple = true
			elseif self.limitionType:match( "split" ) then
				multiple = true
			end
			maxTime = "" .. self.limitValue
		elseif self.limitionType:match( "size" ) then
			if self.limitionType:match( "loop" ) then
				multiple = true
			elseif self.limitionType:match( "multiple" ) then
				multiple = true
			elseif self.limitionType:match( "split" ) then
				multiple = true
			end
			maxSize = "" .. self.limitValue .. "K"
		end

		local filePath = self.location
		if #filePath > 0 and filePath:sub(-1) ~= "/" then
			filePath = filePath .. "/" .. self.sequenceId .. "/"
		else
			filePath = filePath .. self.sequenceId .. "/"
		end

		local muxrate = -1
		if self.fileType == "ts" or self.fileType == "TS" then
			local streams = require "streams"
			muxrate = streams.GetTsMuxrate( self.streamIdx )
		end

		_M.recordGroup:SetOption( self.recordSess, {
			path = filePath,
			type = self.fileType,
			prefix = self.filePrefix or "REC",
			multiple = multiple,
			maxSize = maxSize,
			maxTime = maxTime,
			muxrate = -1,
			ts_pts_pcr_delay = 500,
			ts_null_multiple = 1 --TODO: Here.
		})

		self.lastError = ""
		self.recordingStatus = "preparing"
		self.curSplitStatus = "idle"
		self.curSplitFile = nil
		self.curFileSize = 0
		self.curFileDur = 0

		_M.recordGroup:StartSession( self.recordSess )
		return true
	end

	function OBJ:StopRecord()
		if self.recordingStatus == "recording" or self.recordingStatus == "preparing" then
			if self.recordingStatus == "recording" then
				_M.recordProxy:UpdateRecord( self.sequenceId, "end", self.curSplitFile, self.fileType, os.time(), math.floor(tonumber(self.curFileSize) or 0) )
			end

			_M.recordGroup:StopSession( self.recordSess )
			self.recordSess = nil
			collectgarbage()
			self.recordingStatus = "stopped"
			os.execute( "/bin/sync" ) --NOTE: To write-back the buffered data, to avoid data loss.
			return true
		end
		return true
	end

	function OBJ:Check()
		if self.recordingStatus == "preparing" then
			local status, err = _M.recordGroup:GetSessionStatus( self.recordSess )
			if not status then
				print( "*** Get recording status fail:", err )
				return
			end

			local curSplitStatus, curSplitFile, curFileSize, curFileDur
			curSplitStatus = status.state
			curSplitFile = status.file
			curFileSize = status.size_kb
			curFileDur = status.time

			if curSplitStatus == "recording" then
				self.recordingStatus = "recording"
			else
				return
			end
		end

		if self.recordingStatus == "recording" then
			local status, err = _M.recordGroup:GetSessionStatus( self.recordSess )
			if not status then
				print( "*** Get recording status fail:", err )
				return
			end

			local curSplitStatus, curSplitFile, curFileSize, curFileDur
			curSplitStatus = status.state
			curSplitFile = status.file
			curFileSize = status.size_kb
			curFileDur = status.time
			
			if curSplitFile and curSplitFile ~= "" and curFileSize and curFileDur then
				if self.curSplitFile ~= curSplitFile then
					if self.curSplitFile and self.curSplitFile ~= "" then
						_M.recordProxy:UpdateRecord( self.sequenceId, "end", self.curSplitFile, self.fileType, os.time(), math.floor(tonumber(self.curFileSize) or 0 ) )
					end

					_M.recordProxy:UpdateRecord( self.sequenceId, "new", curSplitFile, self.fileType, os.time(), math.floor(tonumber(curFileSize) or 0 ) )
					self.curSplitFile = curSplitFile
					self.curSplitStatus = curSplitStatus
					self.curFileSize = curFileSize
					self.curFileDur = curFileDur
				end
			end
						
			if curSplitStatus == "recording" then
				local dur = curFileDur - self.curFileDur
				if dur < 0 or dur >= 1 then --NOTE: old dur >= 5
					self.curFileDur = curFileDur
					self.curFileSize = curFileSize
					_M.recordProxy:UpdateRecord( self.sequenceId, "update", self.curSplitFile, self.fileType, os.time(), math.floor(tonumber(curFileSize) or 0) )
				end
			elseif curSplitStatus ~= "preparing" and curSplitStatus ~= "starting" then
				if curSplitFile == self.curSplitFile and curFileSize > 0 and curFileDur > 0 then
					self.curFileDur = curFileDur
					self.curFileSize = curFileSize
				end

				self:StopRecord()
			end
		end
	end

	table.insert( _M.allObjects, OBJ )
	return OBJ
end

----------------------------------------------------
function _M.free(obj)
	local k,v
	for k,v in ipairs( _M.allObjects ) do
		if v == obj then
			v:StopRecord()
			table.remove( _M.allObjects, k )
			break
		end
	end
end

----------------------------------------------------
function _M.Check()
	if _M.recordProxy and not _M.recorderRegistered then
		local mod_streams = require "streams"
		local thisSvc = mod_streams.stream_objs["00"]
		if thisSvc and thisSvc.source.mediaSource then
			print( "*** Media is ready, register recording worker sink to record management service!" )
			_M.recordProxy:RegisterWorker( "any" )
			_M.recorderRegistered = true
		end
	end

	local k,v
	for k,v in ipairs( _M.allObjects ) do
		if v then
			v:Check()
		end
	end
end
----------------------------------------------------

return _M

