local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
require "kmCore"

local _M = {
	each_pusher_session_limits = 4,
	pusher_groups = {
		["00"] = {},
		["01"] = {}
	}
}

----------------------------------------------

local netpusher_meta = {}

function netpusher_meta:AddSession( idr, mediaSource, targetAddr, options )
	if not mediaSource or not targetAddr or targetAddr == "" then
		_dbg.Error( "NET-PUSH", "Invalid network pusher arguments (MediaSource/TargetAddress ...). Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	if self.session then
		_dbg.Warn( "NET-PUSH", "The network pusher session is already created. Stream:", self.stream_id, ", IDR:", idr )
		return self.session
	end

	--Now, parsing the URL to get some arguments.
	local protocol, orig_url_tail
	local user_in_url, pass_in_url
	local ttl_in_url, pkt_size_in_url

	local s

	protocol, s = targetAddr:match( "(%w+)://(.+)$" )
	if not protocol then
		_dbg.Error( "NET-PUSH", "Invalid target URL of:", targetAddr, "(Stream:", self.stream_id, ", IDR:", idr, ")" )
		return nil
	end
	protocol = protocol:lower()
	user_in_url, pass_in_url = s:match( "([^/]-):([^/]*)@.+$" )
	if not user_in_url then
    	local s1 = s:match("(.-)/(.+)")
    	if s1 then
        	user_in_url = s1:match( "([^/]-)@./+$" )
    	end
	end
	if user_in_url then
		orig_url_tail = s:match( ".-@(.+)$" )
	else
		orig_url_tail = s
	end

	if protocol == "udp" or protocol == "rtp" then
		local pfx
		pfx, s = orig_url_tail:match( "(.-)%?(.+)$" )
		if pfx then
			orig_url_tail = pfx
			ttl_in_url = s:match("ttl=(%d+)" )
			ttl_in_url = tonumber(ttl_in_url)
			pkt_size_in_url = s:match("pkt_size=(%d+)")
			pkt_size_in_url = tonumber(pkt_size_in_url)
		end
	end

	if type(options) == "table" then
		local user = options["username"] or options["user"]
		local pass = options["password"] or options["pass"]
		if user and user ~= "" then user_in_url = user end
		if pass then pass_in_url = pass end
		if options["ttl"] then
			ttl_in_url = tonumber(options["ttl"]) or ttl_in_url
		end
		if options["pkt_size"] then
			pkt_size_in_url = tonumber(options["pkt_size"]) or pkt_size_in_url
		end
	else
		options = {
			muxrate = -1
		}
	end

	--Formal the URL
	targetAddr = protocol .. "://"
	if user_in_url then
		targetAddr = targetAddr .. user_in_url
		if pass_in_url then
			targetAddr = targetAddr .. ":" .. pass_in_url
		end
		targetAddr = targetAddr .. "@"
	end

	targetAddr = targetAddr .. orig_url_tail

	if protocol == "udp" or protocol == "rtp" then
		--if ttl_in_url or pkt_size_in_url then
			targetAddr = targetAddr .. "?"
			if ttl_in_url then
				targetAddr = targetAddr .. "ttl=" .. ttl_in_url
			else
				targetAddr = targetAddr .. "ttl=127"
			end
			targetAddr = targetAddr .. "&"
			if pkt_size_in_url and pkt_size_in_url > 0 and pkt_size_in_url <= 1472 then
				targetAddr = targetAddr .. "pkt_size=" .. pkt_size_in_url
			else
				-- RTP限制在6*188+12=1140
				targetAddr = targetAddr .. (protocol == "rtp" and  "pkt_size=1140" or "pkt_size=1316")
			end
		--end
	end

	self.session = self.group.group:AddSession( self.stream_id .. ":" .. (idr or "noname"), mediaSource, 4*1024*1024, 512*1024 )
	if not self.session then
		_dbg.Error( "NET-PUSH", "Fail to create network pusher session for Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	_dbg.Info( "NET-PUSH", "Created network pusher session for Stream:", self.stream_id, ", IDR:", idr, ", URL:", targetAddr )
	if protocol:match("^rtmp") then --That may be: rtmp, rtmps, ...
		protocol = "rtmp"
	end

	self.session_idr = idr
	self.target_url = targetAddr

	options.PushPath = targetAddr
	--options.PushMode = "NONE" --What's means?
	options.PushProtocolType = protocol

	--SRT print
	if protocol == "srt" or protocol == "SRT" then
			_dbg.Info( "NET-PUSH", "Create network SRT pusher session for Stream:",self.target_url )
	end
		
	--For debug
	--local pt = require "pl.pretty"
	--pt.dump(options)

	self.group.group:SetOption( self.session, options )
	return self.session
end

function netpusher_meta:RemoveSession()
	if self.session then
		_dbg.Info( "NET-PUSH", "Removed network pusher session for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.target_url )
		self.group.group:RemoveSession( self.session )
		self.session = nil
		self.session_idr = nil
		self.target_url = nil
		collectgarbage()
	end
	return true
end

function netpusher_meta:StartSession()
	if self.session then
		self.group.group:StartSession( self.session )
		_dbg.Info( "NET-PUSH", "Started network pusher session for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.target_url )
		return true
	else
		_dbg.Error( "NET-PUSH", "No session created while start session, Stream:", self.stream_id )
		return false
	end
end

function netpusher_meta:StopSession()
	if self.session then
		self.group.group:StopSession( self.session )
		_dbg.Info( "NET-PUSH", "Stopped network pusher session for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.target_url )
		return true
	else
		_dbg.Error( "NET-PUSH", "No session created while stop session, Stream:", self.stream_id )
		return false
	end
end

function netpusher_meta:Destroy()
	self:RemoveSession()
	if self.group.counter > 0 then
		self.group.counter = self.group.counter - 1
	end
end

netpusher_meta.__index = netpusher_meta

----------------------------------------------

function _M.GetStreamer( stream_id )
	stream_id = tonumber(stream_id)
	stream_id = string.format( "%02d", stream_id )

	local fetched_group
	local pgroups = _M.pusher_groups[ stream_id ]
	if not pgroups then
		_dbg.Error( "NET-PUSH", "Invalid stream id of:", stream_id )
		return nil
	end

	local i
	for i=1, #pgroups do
		if pgroups[i].counter < _M.each_pusher_session_limits then
			pgroups[i].counter = pgroups[i].counter + 1
			fetched_group = pgroups[i]
			break
		end
	end

	if not fetched_group then
		local idr = "netpush:" .. stream_id .. ":" .. #pgroups + 1
		local net_pusher_group = KM_STREAMING.NetPusher.createNew(idr)
		if not net_pusher_group then
			_dbg.Error( "NET-PUSH", "Fail to create network pusher for stream:", stream_id )
			return nil
		end
		_dbg.Info( "NET-PUSH", "A new network pusher group is created, for stream:", stream_id, ", IDR:", idr )
		fetched_group = {
			counter = 1,
			group = net_pusher_group
		}
		table.insert( pgroups, fetched_group )
	end

	local pusher = {
		stream_id = stream_id,
		group = fetched_group
	}

	return setmetatable( pusher, netpusher_meta )
end

----------------------------------------------
return _M


