local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local timedatectl = require "kiloview.systemd-timedate"
local thread = require "coaf.thread"

require "kmCore"

local _M = {
	each_pusher_session_limits = 4,
	pusher_groups = {
		["00"] = {},
		["01"] = {}
	},
}

----------------------------------------------

local omni_pusher_meta = {}

function omni_pusher_meta:AddSession( idr, mediaSource, targetAddr, options )
	if not mediaSource then
		_dbg.Error( "OMNI-PUSH", "Invalid Sfp pusher arguments (MediaSource). Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	if self.session then
		_dbg.Warn( "OMNI-PUSH", "The Sfp pusher session is already created. Stream:", self.stream_id, ", IDR:", idr )
		return self.session 
	end

	self.session = self.group.group:AddSession( self.stream_id .. ":" .. (idr or "noname"), mediaSource, 4*1024*1024, 512*1024 )
	if not self.session then
		_dbg.Error( "OMNI-PUSH", "Fail to create Sfp pusher session for Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	_dbg.Info( "OMNI-PUSH", "Created Sfp push session, for Stream:", self.stream_id, ", IDR:", idr, ", Target:", targetAddr )
	self.session_idr = idr
	self.targetAddress = targetAddr
	self.savedOptions = options
	self.savedMediaSource = mediaSource

	self.group.group:SetOption( self.session, options )

	return self.session
end

function omni_pusher_meta:RemoveSession()
	if self.session then
		self.group.group:RemoveSession( self.session )
		_dbg.Info( "OMNI-PUSH", "Sfp push session is destroied, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", Target address:", self.targetAddress )
	end

	self.session = nil
	self.session_idr = nil
	self.targetAddress = nil
	self.savedOptions = nil
	self.savedMediaSource = nil
	collectgarbage()

	return true
end

function omni_pusher_meta:StartSession()
	if self.session then
		self.group.group:StartSession( self.session )
		_dbg.Info( "OMNI-PUSH", "Started Sfp pusher session for Stream:", self.stream_id, ", IDR:", self.session_idr, ", Target address:", self.targetAddress )
		return true
	else
		_dbg.Error( "OMNI-PUSH", "No session created while start session, Stream:", self.stream_id )
		return false
	end
end

function omni_pusher_meta:StopSession()
	if self.session then
		self.group.group:StopSession( self.session )
		_dbg.Info( "OMNI-PUSH", "Stopped Sfp pusher session for Stream:", self.stream_id, ", IDR:", self.session_idr, ", Target address:", self.targetAddress )
		return true
	else
		_dbg.Error( "OMNI-PUSH", "No session created while stop session, Stream:", self.stream_id )
		return false
	end
end

function omni_pusher_meta:Destroy()
	self:RemoveSession()
	if self.group.counter > 0 then
		self.group.counter = self.group.counter - 1
	end
end

omni_pusher_meta.__index = omni_pusher_meta

----------------------------------------------

function _M.GetStreamer( stream_id )
	stream_id = tonumber(stream_id)
	stream_id = string.format( "%02d", stream_id )

	local fetched_group
	local pgroups = _M.pusher_groups[ stream_id ]
	if not pgroups then
		_dbg.Error( "OMNI-PUSH", "Invalid stream id of:", stream_id )
		return nil
	end

	local i
	for i=1, #pgroups do
		if pgroups[i].counter < _M.each_pusher_session_limits then
			pgroups[i].counter = pgroups[i].counter + 1
			fetched_group = pgroups[i]
			break
		end
	end

	if not fetched_group then
		local idr = "omni:" .. stream_id .. ":" .. #pgroups + 1
		local omni_pusher_group = KM_STREAMING.sfpPush.createNew( idr )
		if not omni_pusher_group then
			_dbg.Error( "OMNI-PUSH", "Fail to create Sfp pusher for omni-stream:", stream_id )
			return nil
		else
			_dbg.Info( "OMNI-PUSH", "A new Sfp push group '" .. idr .. "' is created." )
		end
		fetched_group = {
			counter = 1,
			group = omni_pusher_group
		}
		table.insert(pgroups, fetched_group)
	end

	local pusher = {
		stream_id = stream_id,
		group = fetched_group
	}

	return setmetatable( pusher, omni_pusher_meta )
end

----------------------------------------------
return _M

