local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
require "kmCore"

local _M = {
	each_pusher_session_limits = 4,
	pusher_groups = {
		["00"] = {},
		["01"] = {}
	},
}

----------------------------------------------
local sproxy_pusher_meta = {}

function sproxy_pusher_meta:AddSession( idr, mediaSource, targetAddr, options )
	if not mediaSource or not targetAddr or targetAddr == "" then
		_dbg.Error( "SPROXY-PUSH", "Invalid SPROXY pusher arguments (MediaSource/TargetAddress ...). Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	if self.session then
		_dbg.Warn( "SPROXY-PUSH", "The SPROXY pusher session is already created. Stream:", self.stream_id, ", IDR:", idr )
		return self.session
	end

	--Now, parsing the URL to get some arguments.
	local protocol, orig_url_tail
	local user_in_url, pass_in_url

	local s

	protocol, s = targetAddr:match( "(%w+)://(.+)$" )
	if not protocol then
		_dbg.Error( "SPROXY-PUSH", "Invalid target URL of:", targetAddr, "(Stream:", self.stream_id, ", IDR:", idr, ")" )
		return nil
	end
	protocol = protocol:lower()
	user_in_url, pass_in_url = s:match( "([^/]-):([^/]*)@.+$" )
	if not user_in_url then
		user_in_url = s:match( "([^/]-)@.+$" )
	end
	if user_in_url then
		orig_url_tail = s:match( ".-@(.+)$" )
	else
		orig_url_tail = s
	end

	if type(options) ~= "table" then
		options = {}
	end

	if user_in_url and not options["user"] then
		options["user"] = user_in_url
	end
	if pass_in_url and not options["pass"] then
		options["pass"] = pass_in_url
	end

	--Formal the URL
	targetAddr = protocol .. "://" .. orig_url_tail
	
	self.session = self.group.group:AddSession( mediaSource, targetAddr, options )
	if not self.session then
		_dbg.Error( "SPROXY-PUSH", "Fail to create SPROXY pusher session for Stream:", self.stream_id, ", IDR:", idr )
		return nil
	end

	_dbg.Info( "SPROXY-PUSH", "Created SPROXY push session, for Stream:", self.stream_id, ", IDR:", idr, ", URL:", targetAddr )
	self.session_idr = idr
	self.formalUrl = targetAddr
	self.savedOptions = options
	self.savedMediaSource = mediaSource

	return self.session
end

function sproxy_pusher_meta:RemoveSession()
	if self.session then
		self.group.group:RemoveSession( self.session )
		_dbg.Info( "SPROXY-PUSH", "SPROXY push session is destroied, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
		self.session = nil
		self.session_idr = nil
		self.formalUrl = nil
		self.savedOptions = nil
		self.savedMediaSource = nil
		collectgarbage()
	end
	return true
end

function sproxy_pusher_meta:StartSession()
	if self.session then
		return true
	end
	if not self.formalUrl or not self.savedMediaSource then
		_dbg.Warn( "SPROXY-PUSH", "No session created for Stream:", self.stream_id )
		return false
	end

	self.session = self.group.group:AddSession( self.savedMediaSource, self.formalUrl, self.savedOptions )
	if not self.session then
		_dbg.Error( "SPROXY-PUSH", "Fail to create SPROXY pusher session for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
		return false
	end
	_dbg.Info( "SPROXY-PUSH", "SPROXY push session is started, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
	return true
end

function sproxy_pusher_meta:StopSession()
	if self.session then
		self.group.group:RemoveSession( self.session )
		_dbg.Info( "SPROXY-PUSH", "SPROXY push session is stopped, for Stream:", self.stream_id, ", IDR:", self.session_idr, ", URL:", self.formalUrl )
		self.session = nil
		collectgarbage()
	end
	return true
end

function sproxy_pusher_meta:Destroy()
	self:RemoveSession()
	if self.group.counter > 0 then
		self.group.counter = self.group.counter - 1
	end
end

sproxy_pusher_meta.__index = sproxy_pusher_meta

----------------------------------------------

function _M.GetStreamer( stream_id )
	stream_id = tonumber(stream_id)
	stream_id = string.format( "%02d", stream_id )

	local fetched_group
	local pgroups = _M.pusher_groups[ stream_id ]
	if not pgroups then
		_dbg.Error( "SPROXY-PUSH", "Invalid stream id of:", stream_id )
		return nil
	end

	local i
	for i=1, #pgroups do
		if pgroups[i].counter < _M.each_pusher_session_limits then
			pgroups[i].counter = pgroups[i].counter + 1
			fetched_group = pgroups[i]
			break
		end
	end

	if not fetched_group then
		local idr = "sproxy:" .. stream_id .. ":" .. #pgroups + 1
		local sproxy_pusher_group = KM_STREAMING.Sproxy.createNew(idr)
		if not sproxy_pusher_group then
			_dbg.Error( "SPROXY-PUSH", "Fail to create SPROXY pusher for stream:", stream_id )
			return nil
		else
			_dbg.Info( "SPROXY-PUSH", "A new SPROXY push group '" .. idr .. "' is created." )
		end
		fetched_group = {
			counter = 1,
			group = sproxy_pusher_group
		}
		table.insert(pgroups, fetched_group)
	end

	local pusher = {
		stream_id = stream_id,
		group = fetched_group
	}

	return setmetatable( pusher, sproxy_pusher_meta )
end

----------------------------------------------
return _M


