local rpc = require "kiloview.ringRpc"
local nu = require "kiloview.netUtil"

local _M = {}

--Get default ethernet device.
function _M.GetDefaultEthernet(objId)
	local r, devices = rpc.Call( "NETWORK_MANAGER.GetDevicesBy", "*", "ETHERNET" )
	if not r or not devices or #devices < 1 then
		return nil
	elseif not objId or objId == "" or objId == "/" then
		return devices[1]
	else
		local k,v
		for k,v in pairs(devices) do
			if v.Object == objId then
				return v
			end
		end
		return nil
	end
end

--Get ethernet detail
function _M.GetEthernetDetails( device )
	if not device then
		return nil
	end

	local localLinkIP = ""
	local r,cc,gets = rpc.Call( "BUS.Get", { ["hostapm.localLink.ip"] = "", ["hostapm.localLink.enable"] = "true" } )
	if r and gets and gets[ "hostapm.localLink.enable" ] == "true" then
		localLinkIP = gets["hostapm.localLink.ip"]
	end

	local rslt = {}
	
	local r,devInfo = rpc.Call( "NETWORK_MANAGER.GetDeviceInfo", device.Object, { WIRE = true, IP4 = true } )
	if devInfo and devInfo.WIRE and devInfo.WIRE.Carrier and devInfo.WIRE.Carrier ~= 0 then
		rslt.Carrier = 1
	else
		rslt.Carrier = 0
	end

    local specConn

    --TODO: Check here, perhaps no need this.
	if devInfo.ActiveConnection and devInfo.ActiveConnection ~= "" and devInfo.ActiveConnection ~= "/" then
		local r, ac = rpc.Call( "NETWORK_MANAGER.GetActivateConnectionDetails", devInfo.ActiveConnection )
		if r and ac then
			specConn = ac[devInfo.ActiveConnection].Connection
            --Test only.
            --print( "***", specConn, devInfo.ActiveConnection )
		end
	end

	if not specConn then
		specConn = devInfo.AvailableConnections[1]
		if not specConn then
			local r, allConns = rpc.Call( "NETWORK_MANAGER.ListConnectionAndGetDetails", { interface=devInfo.Name } )
			if r and allConns then
				local k,v
				for k,v in pairs(allConns) do
					specConn = k	--Get first one.
					break
				end
			end
		end
	end

    --Get ethernet connection details.
	if specConn then
        rslt.ConnectionObject= specConn

		local r, connInfo = rpc.Call( "NETWORK_MANAGER.GetConnectionDetails", specConn )

        if r and connInfo and connInfo[specConn] then
		    connInfo = connInfo[specConn]
		    rslt.ConnectionUuid = connInfo["connection"]["uuid"]
		    rslt.DHCP = connInfo["ipv4"]["method"] and connInfo["ipv4"]["method"] == "auto" or false

			if connInfo["ipv4"]["dns"] and #connInfo["ipv4"]["dns"] > 0 then
				rslt.Dns1 = connInfo["ipv4"]["dns"][1] or ""
				rslt.Dns2 = connInfo["ipv4"]["dns"][2] or ""
            else
                rslt.Dns1 = ""
                rslt.Dns2 = ""
			end
        else
            rslt.ConnectionUuid= ""
            rslt.DHCP = true
            rslt.Dns1 = ""
            rslt.Dns2 = ""
        end
    else
        rslt.ConnectionObject = ""
        rslt.ConnectionUuid= ""
        rslt.DHCP = true
        rslt.Dns1 = ""
        rslt.Dns2 = ""
    end


    rslt.Speed = devInfo.WIRE.Speed or 0
    rslt.HwAddress = devInfo.WIRE.HwAddress or ""
    rslt.PermHwAddress= devInfo.WIRE.PermHwAddress or ""

	if devInfo.IP4 and devInfo.IP4.Addresses and #devInfo.IP4.Addresses > 0 then
        local i
        for i=1, #devInfo.IP4.Addresses do
            if devInfo.IP4.Addresses[i].address ~= localLinkIP then
		        rslt.IP = devInfo.IP4.Addresses[i].address
		        rslt.Prefix = devInfo.IP4.Addresses[i].prefix
		        rslt.Gateway = devInfo.IP4.Gateway
                return rslt
            end
        end

        --No other valid address, so use the default too.
		rslt.IP = devInfo.IP4.Addresses[1].address
		rslt.Prefix = devInfo.IP4.Addresses[1].prefix
		rslt.Gateway = devInfo.IP4.Gateway
	else
		rslt.IP = ""
		rslt.Prefix = 0
		rslt.Gateway = ""
	end

	return rslt
end

--Set ethernet configure
function _M.SetConnection( deviceObj, new, args )
	local r, devInfo = rpc.Call( "NETWORK_MANAGER.GetDeviceInfo", deviceObj, {WIRE = true} )	--Get WIRE informations, so I can get MAC.
	if not r or not devInfo then
		return false, "ERROR: Fail to get ethernet device informations"
	end

	local devName = devInfo.Name

	local connSettings = {
		["connection"] = {},
		["802-3-ethernet"] = {},
		["ipv4"] = {}
	}

	connSettings["connection"]["type"] = "802-3-ethernet"
	connSettings["connection"]["autoconnect"] = true	--Force auto connection.

	if args.Uuid and args.Uuid ~= "" then
		connSettings["connection"]["uuid"] = args.Uuid
		local r, detail = rpc.Call( "NETWORK_MANAGER.ListConnectionAndGetDetails", {uuid=args.Uuid} )
		if r and detail then
			local k,v
			local found
			for k,v in pairs(detail) do
				found = v
				break
			end
			
			if found then
				connSettings["connection"]["id"] = found["connection"]["id"]
			end
		end
	end

	if not connSettings["connection"]["id"] then
		connSettings["connection"]["id"] = devName
	end
	
	if args.CloneMac then
		if args.CloneMac ~= "" then
			connSettings["802-3-ethernet"]["cloned-mac-address"] = nu.MACToArray(args.CloneMac)
		--else
		--	connSettings["802-3-ethernet"]["cloned-mac-address"] = nu.MACToArray(devInfo.WIRE.PermHwAddress)
		end
	end

	connSettings["connection"]["interface-name"] = devName

	if args.DHCP and tonumber(args.DHCP) == 1 then	 --DHCP
		connSettings["ipv4"]["method"] = "auto"
	else
		connSettings["ipv4"]["method"] = "manual"

		if args.IP and args.IP ~= "" then
			connSettings["ipv4"]["address-data"] = {{address=args.IP, prefix= tonumber(args.Prefix) or nu.ip4MaskToPrefix(args.Mask or "255.255.255.0")}}
		end

		connSettings["ipv4"]["gateway"] = args.Gateway or ""

		connSettings["ipv4"]["dns"] = { args.Dns1 and args.Dns1 ~= "" and args.Dns1 or nil, args.Dns2 and args.Dns2 ~= "" and args.Dns2 or nil }
		if #connSettings["ipv4"]["dns"] < 1 then
			connSettings["ipv4"]["dns"] = nil
		end
	end

	if new then
		local r, o = rpc.Call( "NETWORK_MANAGER.AddConnection", connSettings )
		if r and o then
			return true
		else
			return false, "Add new ethernet connection fail"
		end
	else
		local r, o = rpc.Call( "NETWORK_MANAGER.ChangeConnectionDetails", args.Object, connSettings )
		if r and o then
			return true
		else
			--print( r, o )
			return false, "Change ethernet connection settings fail"
		end
	end
end


--(re)connect ethernet connection
function _M.ConnectConnection( deviceObj )
	local r, o = rpc.Call( "NETWORK_MANAGER.ConnectDevice", deviceObj )
	if r and o then
		return true
	else
		return false, "Fail to (re)connect ethernet connection"
	end
end
----------------------------------------------------------------------------------

return _M
