------------------------------------------------------
-- KILOVIEW Device HTTP Api, v1.0
-- Copyright (c) 2016, KILOVIEW Electronics.
------------------------------------------------------

--[[

##ABSTRACT{G=ENCODING}: 获取音频编码引擎的通道列表
>>
>> KILOVIEW编码器设备有一个独立的音频编码引擎。这个引擎可以支持1个或多个通道。每个通道可以
从不同的音频输入设备采集音频，并压缩成不同的格式。对于一个视频编码码流，它可以绑定到音频
编码引擎上的指定通道，从而实现视频和音频的合并码流。

##ARGUMENTS{GET,POST}: 
    @@Channel{INT}: 要获取音频通道的ID，可选。如果指定了Channel，则返回值中只有该Channel的内容。否则将列出所有的Channel。

    @@Config{INT}: 
        可选。
        %%1: 指明需要获取通道参数配置。（参见 返回值 中的"ChannelConfigs"）
        %%0: (或不指定[默认]) 不返回通道参数配置。
    ~~

    @@Status{INT}: 
        可选。
        %%1: 指明需要获取通道编码状态（不指定Status参数时，默认该选项）。（参见 返回值 中的"ChannelStatus"）
        %%0: 不返回通道编码状态。
    ~~

##RETURNS: 
	++Data{OBJ}: 返回结果对象
        ++ChannelConfigs{ARR}: 数组，每一项表示一个音频编码通道的配置，格式如下：
            ++{OBJ}:
                **ID{NUM}:  通道ID
                **Enable{INT}: 
                    %%0: 表示该通道被禁用
                    %%1: 表示启用
                ~~
                **Device: 音频采集设备名称，形式如 alsa:default。目前请忽略该值的作用。
                **Codec: 音频编码算法，目前支持AAC(Low-Cost)和G711算法。
                    %%AAC: AAC(Low-Coast)
                    %%G711: G.711
                ~~

                **Sampling{INT}: 采样率。对AAC有效，对G711算法，恒为8KHz
                **Channels{INT}: 声道，值为1或2。对AAC有效，对G711算法，恒为1
                **Bitrate{NUM}: 编码码率，单位bps（如64000表示64Kbps）。对AAC有效，对G711算法，恒为64000
                **AAC_Format: AAC格式。
                    %%RAW: 表示原生的AAC格式；
                    %%ADTS: 表示带有ADTS头。
                            >>{BOLD} 对于TS流应用，必须指定AAC_Format为ADTS
                ~~

                **G711_Format: G.711格式。
                    %%ULAW: G.711 uLaw格式
                    %%ALAW: G.711 aLaw格式
                ~~
            ~~
        ~~

        ++ChannelStatus{ARR}: 数组，每一项表示一个音频编码通道的状态，格式如下：
            ++{OBJ}:
                **ID{INT}: 数值, //通道ID
                **Enable{INT}: 
                    %%0: 表示该通道被禁用
                    %%1: 表示启用
                ~~
                **Codec: 音频编码算法，目前支持AAC(Low-Cost)和G711算法。
                    %%AAC: AAC(Low-Coast)
                    %%G711: G.711
                ~~
                **Sampling{INT}: 采样率。对AAC有效，对G711算法，恒为8KHz
                **Channels{INT}: 声道，值为1或2。对AAC有效，对G711算法，恒为1
                **Bitrate{NUM}: 实时编码码率，单位bps（如64000表示64Kbps）。对AAC有效，对G711算法，恒为64000
                **AAC_Format: AAC格式。
                    %%RAW: 表示原生的AAC格式；
                    %%ADTS: 表示带有ADTS头。
                            >>{BOLD} 对于TS流应用，必须指定AAC_Format为ADTS
                ~~

                **G711_Format: G.711格式。
                    %%ULAW: G.711 uLaw格式
                    %%ALAW: G.711 aLaw格式
                ~~
            ~~
        ~~
	~~
]]

local cgilua = require "cgilua"
local rpc = require "kiloview.ringRpc"

local specChannel = nil
local retConfig = false
local retStatus = true

local args = require("M_Request").VerifyAndGetArgs()
if args then
	if args.Channel then
		specChannel = tonumber(args.Channel)
	end
	if args.Config and tonumber(args.Config) == 1 then
		retConfig = true
	end
	if args.Status and tonumber(args.Status) == 0 then
		retStatus = false
	end
end

local maxAEngChannels = 1

local SAPI = SAPI or cgilua.SAPI
SAPI.Response.contenttype( "application/json; charset=utf-8" )

if specChannel and ( specChannel <= 0 or specChannel > maxAEngChannels ) then
	cgilua.put([[
{
	"Result":400,
	"Status":"ERROR: Invalid audio engine channel"
}
]])
	return
end

cgilua.put([[
{
	"Result":200,
	"Status":"OK",
	"Data": {
]])

local statusStr = ""
local i
for i=1, maxAEngChannels do
	if not specChannel or specChannel == i then
		local thisChannel = "audioEngine.encoders[" .. i .. "]"
		local gets = {
			[ thisChannel .. ".enable" ] = 0,
			[ thisChannel .. ".source.type" ] = "ALSA",
			[ thisChannel .. ".source.device" ] = "default",
			[ thisChannel .. ".encoder.codec" ] = "AAC",
			[ thisChannel .. ".encoder.sampling" ] = 48000,
			[ thisChannel .. ".encoder.channels" ] = 2,
			[ thisChannel .. ".encoder.bitrate" ] = 64000,
			[ thisChannel .. ".encoder.aac.type" ] = "AAC-LC",
			[ thisChannel .. ".encoder.aac.format" ] = "RAW",
			[ thisChannel .. ".encoder.g711.law" ] = "ULAW"
		}
		local r,done,rslt = rpc.Call( "BUS.Get", gets )

		if i > 1 then
			statusStr = statusStr .. ","
		end

		if not r or not rslt then
			statusStr = statusStr .. [[
		{
			"ID":]] .. i .. [[,
			"Enable":0,
			"ERROR":1
		}]]
		else
			statusStr = statusStr .. [[
		{
			"ID":]] .. i .. [[,
			"Enable":]] .. ( i==1 and 1 or gets[thisChannel .. ".enable"]) .. [[,
			"Device":"]] .. gets[thisChannel .. ".source.device"] .. [[",
			"Codec":"]] .. gets[thisChannel .. ".encoder.codec"] .. [[",
			"Sampling":]] .. gets[thisChannel .. ".encoder.sampling"] .. [[,
			"Channels":]] .. gets[thisChannel .. ".encoder.channels"] .. [[,
			"Bitrate":]] .. gets[thisChannel .. ".encoder.bitrate"] .. [[,
			"AAC_Format":"]] .. gets[thisChannel .. ".encoder.aac.format"] .. [[",
			"G711_Format":"]] .. gets[thisChannel .. ".encoder.g711.law"] .. [["
		}]]
		end
	end
end

if retConfig then
	cgilua.put( [[
		"ChannelConfigs":[
	]] )

	cgilua.put( statusStr )

	cgilua.put( [[
		]
	]] )
end

if retStatus then
	if retConfig then
		cgilua.put( [[,
	]] )
	end

	cgilua.put([[
		"ChannelStatus":[
]])
	cgilua.put( statusStr )
	cgilua.put( [[
		]
	]] )
end

cgilua.put([[
	}
}
]])

