------------------------------------------------------
-- KILOVIEW Device HTTP Api, v1.0
-- Copyright (c) 2016, KILOVIEW Electronics.
------------------------------------------------------

--[[

##ABSTRACT{G=ENCODING}: 获得当前MotionJPEG编码和Streaming参数

>>    MotionJPEG编码/Streaming在所有KILOVIEW编码产品中均是默认开启的，以辅助于视频抓拍
图像用于预览。但MotionJPEG可以配置成不同的模式，如仅用于预览（图像较小、帧率5fps），或
与主码流相同，或与子码流相同，以供不同的应用场景。

##ARGUMENTS{GET,POST}: 
    @@Config: 可选参数。
        %%1: 指定要获取MotionJPEG的编码配置参数（参见返回值中的 "Config"）
        %%0: （或不指定）[默认] 不获取配置参数
    ~~

    @@Status: 可选参数。
        %%1: 指定要获取MotionJPEG的实时状态信息（参见返回值中的 "Status"）
        %%0: （或不指定）[默认] 不获取状态信息
    ~~

##RETURNS: 
	++Data{OBJ}:
        **Type: 当前MotionJPEG的配置模式
            %%preview: 仅预览模式（720x405大小，5fps）
            %%main: 图像大小与主码流相同（帧率可设置）
            %%sub: 图像大小与子码流相同（帧率可设置）
        ~~
        **SnapUrl: 
            http://..., 
            >> 用于获取预览视频图片的HTTP URL
        
        ++Config{OBJ}:   当前MotionJPEG的编码配置参数 (请求参数中如果指定Config=1返回)
            **Quality{INT}: 1-99, 指定JPEG的编码图像质量
            **Chrome{INT}: 0|1, 0表示彩色，1表示灰度

            **FrameRateMode: 编码帧率模式
                %%default: 表示全帧率，即输入视频帧率为多少，编码帧率为多少
                %%half: 表示当输入视频的帧率 >=50 时，自动降半帧率
                %%custom: 表示手动指定帧率
            ~~

            **FrameRate{NUM}: 数值, 指定的编码帧率（如果FrameRateMode为full/half，则不必要指定FrameRate值）

            **BindAudioChannel{INT}: 数值, 绑定的音频编码引擎通道号, -1表示无音频绑定
        ~~
        ~~
        ++Status{OBJ}: 当前实时编码状态信息 (请求参数中如果指定Status=1返回)
            **BindAudioChannel{INT}: 绑定的音频编码引擎通道号, -1表示无音频绑定（同Config中的同名参数）
            **Width{INT}: 实际编码的视频宽度
            **Height{INT}: 实际编码的视频高度
            **FrameRate{NUM}: 实际编码的帧率(fps)
            **Bitrate{INT}: 视频实际编码的实时码率(Kbps)
        ~~
	~~
##NOTES:
    注：音频如果需要获取实时状态，请根据BindAudioChannel参数的ID，另行获取音频的实时编码状态，参考 ^^REF{getAudioEngChannel.lua}
]]

local cgilua = require "cgilua"
local rpc = require "kiloview.ringRpc"

local retConfig = false
local retStatus = false

local args = require("M_Request").VerifyAndGetArgs()
if args then
	if args.Config and tonumber(args.Config) == 1 then
		retConfig = true
	end
	if args.Status and tonumber(args.Status) == 1 then
		retStatus = true
	end
end

local host = cgilua.servervariable"SERVER_ADDR"
local port = 80
local r,cc,rets = rpc.Call( "BUS.Get", { ["webConfig.port"] = 80 } )
if r and rets then
	port = rets["webConfig.port"]
end

local argPrefix = "encoder.stream[2]"
local reqParams = {
	["streamer.stream[2].audioChannel"] = -1,
	["encoder.snap_type"] = "preview",
--	["streamer.kmp_ordering"] = "default",
	[argPrefix .. ".video.scale"] = "default",
	[argPrefix .. ".video.framerate_mode"] = "default",
	[argPrefix .. ".video.framerate"] = 0,
	[argPrefix .. ".video.quality"] = 80,
	[argPrefix .. ".video.chrome"] = 1
}

local SAPI = SAPI or cgilua.SAPI
SAPI.Response.contenttype( "application/json; charset=utf-8" )

r, cc, rets = rpc.Call( "BUS.Get", reqParams )
if not r or not rets then
	cgilua.put([[
{
	"Result":400,
	"Status":"ERROR: Fail to get Motion-JPEG stream configurations"
}
]])
	return
end

cgilua.put([[
{
	"Result":200,
	"Status":"OK",
	"Data":{
		"Type":"]] .. rets["encoder.snap_type"] .. [[",
		"SnapUrl":"http://]] .. host .. ":" .. port .. [[/actions/snap.lua"
]])

if retConfig then
cgilua.put([[
		,"Config":{
			"Quality":]] .. rets[argPrefix .. ".video.quality"] .. [[,
			"Chrome":]] .. rets[argPrefix .. ".video.chrome"] .. [[,
			"FrameRateMode":"]] .. rets[argPrefix .. ".video.framerate_mode"] .. [[",
			"FrameRate":]] .. rets[argPrefix .. ".video.framerate"] .. [[,
			"BindAudioChannel":]] .. rets["streamer.stream[2].audioChannel"] .. [[
		}
]])
end

if retStatus then
	local x, cfgs = rpc.Call( "BUS.GetStreamStatus", "02" )
	if not x or not cfgs then
		cgilua.put([[
		,"Status":{
			"BindAudioChannel":]] .. rets["streamer.stream[2].audioChannel"] .. [[
		}
]])
	else
		cgilua.put([[
		,"Status":{
			"BindAudioChannel":]] .. rets["streamer.stream[2].audioChannel"] .. [[,
			"Width":]] .. (cfgs["video.width"] or 0) .. [[,
			"Height":]] .. (cfgs["video.height"] or 0) .. [[,
			"FrameRate":]] .. (cfgs["video.frameRate"] or 0) .. [[,
			"Bitrate":]] .. (cfgs["video.bitrate"] or 0) .. [[
		}
]])
	end
end

cgilua.put([[
	}
}]])
