#!/usr/bin/env lua
local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local thread = require "coaf.thread"
local service = require "coaf.service"

local socket = require "cosocket.core"
local roll = require "coaf.log.rollfile"

local OSD_IMAGE_LOC = "/data/osd_images"
local lfs = require "lfs"
local pretty = require "pl.pretty"

-- Change current dir to app location
coaf.ChDir(coaf.GetAppDir(arg))

------------------------------------------------
local withLog = false
local _x=1
while _x <= #arg do
	if arg[_x] == "--log" then
		withLog = true
	end
	_x = _x+1
end

-- First, initialize the log file.
if withLog then
	os.execute( "/bin/mkdir -p /var/log" )
	local logF = roll.Create("/var/log/img_manager.log")
	if not logF or not logF:init() then
		_dbg.Error("IMAGE", "Fail to open log file: /var/log/img_manager.log, so the log messages should be output to console.")
	else
		_dbg.AddLogger("logfile", logF, nil, nil)
	end
end


------------------------------------------------
_G.g_service = nil
------------------------------------------------

local img = {
}

local function SaveDesc( id, cfgFile, tb )
	local f = io.open( cfgFile, "w" )
	if not f then
		return false, "Fail to create image description, for image ID:" .. id
	end
	
	local cont = pretty.write(tb, "\t")
	f:write( "return {\n" )
	f:write( cont )
	f:write( "\n}\n" )
	f:close()
	return true
end

function img.ConvertImage(src_file, dst_file, size_w, size_h, alphaThres, tp_x, tp_y, tp_mode, fuzz )
	if not size_w or not size_h or size_w <= 0 or size_h <= 0 then
		local f = io.popen( "/usr/bin/identify '" .. src_file .. "'", "r" )
		if f then
			local line = f:read("*l")
			f:close()
			if line then
				size_w, size_h = line:match(".-([0-9]+)[xX]([0-9]+)")
				size_w = tonumber(size_w)
				size_h = tonumber(size_h)
				if not size_w or not size_h then
					return false
				end
			else
				return false
			end
		else
			return false
		end
	end

	--local restrict_w = (size_w + 31) & 0xffffffe0
	--local restrict_w = (size_w + 15) & 0xfffffff0
	
	local restrict_w = size_w
	local restrict_h = (size_h + 3) & 0xfffffffc

	alphaThres = tonumber(alphaThres) or 0
	tp_mode = tp_mode or "replace"
	fuzz = tonumber(fuzz) or 5

	local cmd = "/usr/bin/convert '" .. src_file .. "' -resize " .. size_w .. "x" .. size_h .. 
				" -alpha set -background none"
	
	if alphaThres > 0 then
		cmd = cmd .. " -channel alpha -threshold " .. alphaThres .. "%"
	end

	if tp_x and tp_y then
		cmd = cmd .. " -fill none -fuzz " .. fuzz .. "% -draw 'color " .. tp_x .. "," .. tp_y .. " " .. tp_mode .. "'"
	end

	if restrict_w >= restrict_h then
		cmd = cmd .. " -gravity center -extent " .. restrict_w .. "x" .. restrict_h .. " -depth 8 -colors 255 " .. " ycbcra:- | /usr/bin/y2raw -p 2103 -o '" .. dst_file .. "'"
	else --A fuck imagemagick bug.
		cmd = cmd .. " -gravity center -extent " .. restrict_h .. "x" .. restrict_h .. " -depth 8 -colors 255 " .. 
			  " ycbcra:- | /usr/bin/y2raw -p 2103 -c " .. restrict_h .. ":" .. restrict_h .. ":" .. restrict_w .. ":" .. restrict_h .. " -o '" .. dst_file .. "'"
	end

	local r,stat,code = os.execute( cmd )
	if r and code == 0 then
		return true, size_w, size_h, restrict_w, restrict_h
	else
		return false, size_w, size_h, restrict_w, restrict_h
	end
end

function img.GetImageDesc(id)
	if type(id) ~= "string" then
		return nil
	end
	local cfgFile = OSD_IMAGE_LOC .. "/" .. id .. "/images.desc"
	local x, ret = pcall( dofile, cfgFile )
	if x and type(ret) == "table" and type(ret[1]) == "table" then
		return ret[1]
	else
		return nil
	end
end

function img.ResizeImage(id, w, h, a, tp_x, tp_y, tp_mode, fuzz)
	w = tonumber(w)
	h = tonumber(h)
	if w and w <= 0 then
		w = nil
	end

	if h and h <= 0 then
		h = nil
	end

	tp_x = tonumber(tp_x)
	tp_y = tonumber(tp_y)
	if tp_x and tp_x < 0 then tp_x = nil end
	if tp_y and tp_y < 0 then tp_y = nil end

	fuzz = tonumber(fuzz)

	local cfgFile = OSD_IMAGE_LOC .. "/" .. id .. "/images.desc"
	local x, ret = pcall( dofile, cfgFile )
	if x and type(ret) == "table" and type(ret[1]) == "table" then
		ret = ret[1]
		local act_w, act_h, res_w, res_h
		x, act_w, act_h, res_w, res_h = img.ConvertImage( OSD_IMAGE_LOC .. "/" .. id .. "/" .. ret.orig, OSD_IMAGE_LOC .. "/" .. id .. "/default.raw", w, h, a, tp_x, tp_y, tp_mode, fuzz )
		if x then
			if tp_x and tp_y then
				ret.transparent = { tp_x, tp_y, tp_mode or "replace", fuzz or 5 }
			else
				ret.transparent = nil
			end
			ret.png_alpha_thres = a
			ret.size = { act_w, act_h, res_w, res_h }
			SaveDesc( id, cfgFile, ret )
			return true
		else
			return false, "Fail to convert image"
		end
	else
		return false, "Invalid image"
	end
end

function img.SetAlpha(id,isAlpha)
	isAlpha = tonumber(isAlpha)
	if isAlpha and (isAlpha < 0 or isAlpha > 255) then
		isAlpha = 255
	end

	local cfgFile = OSD_IMAGE_LOC .. "/" .. id .. "/images.desc"
	local r, ret = pcall( dofile, cfgFile )
	if r and type(ret) == "table" and type(ret[1]) == "table" then
			ret = ret[1]
			ret.alpha = isAlpha
			SaveDesc( id, cfgFile, ret )
			return true
	else
		return false, "Invalid image"
	end
end

function img.GetList()
	local rslt = {}
	local item, attrs
	local counter = 0
	attr = lfs.attributes( OSD_IMAGE_LOC )
	if not attr or attr.mode ~= "directory" then
		return rslt
	end
	for item in lfs.dir( OSD_IMAGE_LOC ) do
		if item ~= "." and item ~= ".." then
			attrs = lfs.attributes( OSD_IMAGE_LOC .. "/" .. item )
			if attrs and attrs.mode == "directory" then
				local cfgFile = OSD_IMAGE_LOC .. "/" .. item .. "/images.desc"
				local x, ret = pcall( dofile, cfgFile )
				if x and type(ret) == "table" and type(ret[1]) == "table" then
					counter=counter + 1
					ret[1].id = item
					table.insert( rslt, ret[1] )
				end
			end
		end
	end
	table.sort( rslt, function(a,b) return a.id < b.id end )
	return counter, rslt
end

function img.Add( id, file )
	if not id or id == "" or not file or file == "" then
		return false, "Invalid IMAGE ID or file"
	end

	local attrs = lfs.attributes( OSD_IMAGE_LOC .. "/" .. id .. "/images.desc" )
	if attrs then
		return false, "IMAGE ID is already exist"
	end

	local filetype = file:match( "^.+%.(.*)$" )
	if not filetype or (filetype:lower() ~= "jpg" and filetype:lower() ~= "jpeg" and filetype:lower() ~= "png") then
		return false, "Invalid image format, only JPEG/PNG support"
	end

	os.execute( "/bin/rm -Rf '" .. OSD_IMAGE_LOC .. "/" .. id .. "'" )
	os.execute( "/bin/mkdir -p '" .. OSD_IMAGE_LOC .. "/" .. id .. "'" )
	local r,stat,code
	r,stat,code = os.execute( "/bin/cp -Rf '" .. file .. "' '" .. OSD_IMAGE_LOC .. "/" .. id .. "/orig." .. filetype .. "'" )
	if not r or code ~= 0 then
		os.execute( "/bin/rm -Rf '" .. OSD_IMAGE_LOC .. "/" .. id .. "'" )
		return false, "Fail to save image, for image ID:" .. id
	end

	local orig_w, orig_h, res_w, res_h
	r, orig_w, orig_h, res_w, res_h = img.ConvertImage( file, OSD_IMAGE_LOC .. "/" .. id .. "/default.raw" )
	if not r then
		os.execute( "/bin/rm -Rf '" .. OSD_IMAGE_LOC .. "/" .. id .. "'" )
		return false, "Fail to convert image, for image ID:" .. id
	end

	local f = io.open( OSD_IMAGE_LOC .. "/" .. id .. "/images.desc", "w" )
	if not f then
		os.execute( "/bin/rm -Rf '" .. OSD_IMAGE_LOC .. "/" .. id .. "'" )
		return false, "Fail to create image description, for image ID:" .. id
	end

	f:write([[
return {
	{
		type = "]] .. filetype .. [[",
		orig = "orig.]] .. filetype .. [[",
		orig_size = "]] .. orig_w .. "x" .. orig_h .. [[",
		file = "default.raw",
		alpha = nil,
		png_alpha_thres = nil,
		transparent = nil,
		size = {]] .. orig_w .. "," .. orig_h .. "," .. res_w .. "," .. res_h .. [[}
	}
}
	]])
	
	f:close()
	return true
end

function img.Remove( id )
	if not id or id == "" then
		return false, "Invalid IMAGE ID"
	end
	os.execute( "/bin/rm -Rf '" .. OSD_IMAGE_LOC .. "/" .. id .. "'" )
	return true
end

------------------------------------------------

coaf.Init( "img_manager" )
-- Initialize service.
service.InitSocket()

---------------------------------------------------------

local function Main(args)
	g_service = service.New( "img_manager", "unix:/var/run/img.sock" )
	if g_service then
		g_service:CreateCoServerProxy()
		g_service:ExposeObject("/", img )
	else
		_dbg.Error( "IMAGE", "Fail to create CoService for img_manager" )
	end
end

coaf.Start(Main,arg)

_dbg.Print( "IMAGE", "img_manager service is terminated." )

