local presetM = require "kiloview.presetManager"
local getset = require "kiloview.getSetHelper"
local nu = require "kiloview.netUtil"
local _dbg = require "coaf.debug"
local client = require "coaf.client"

local _M = {}

require "features"
local x, reset = pcall( require, "kiloview." .. _G.DEVICE_TYPE .. ".reset" )
if not x or not reset then
	x, reset = pcall( require, "kiloview.reset" )
end

-----------------------------------------
local function GetCoAFProxy( objPath, service, loc, timeout )
	local proxy
	local cli
	cli = client.New( service, loc or "local", nil, timeout or 3 ) --3 seconds timeout.
	if not cli then
		_dbg.Error( "BUS_OPR", "Fail to create connection to service: " .. (loc or "local") .. "/" .. (service or "nil") )
		return nil
	end
	proxy = cli:CreateProxy(objPath, true) --Exclusive
	if not proxy then
		_dbg.Error( "BUS_OPR", "Fail to create object proxy for '" .. (objPath or "unknown") .. "'" )
		return nil
	end
	return proxy
end

-----------------------------------------

local NETWORK_MANAGER = {}
local nm_meta = {}
function nm_meta:__index(func)
	local invoker = function(...)
		local proxy = GetCoAFProxy( "/raw/NETWORK_MANAGER", "nm_daemon", "local", 5 )
		if not proxy then
			_dbg.Error( "NETWORK_MANAGER", "Fail to get network manager proxy" )
			return nil
		end
		local rets = { proxy[func](...) }
		proxy:Destroy()
		if rets and rets[1] then
			table.remove(rets,1)
			return table.unpack(rets)
		else
			_dbg.Error( "NETWORK_MANAGER", "Network manager invoke '" .. func .. "' error:", rets and rets[2] )
			return nil
		end
	end
	NETWORK_MANAGER[func] = invoker
	return invoker
end

function NETWORK_MANAGER.Init()
	return true
end

setmetatable( NETWORK_MANAGER, nm_meta)

-----------------------------------------
--- Get/Set operations.
-----------------------------------------

local getSetTargets = {
	["encoder"] = { bus = "codec", obj = "/Codec/Encoders" },
	["streamer_global"] = { bus = "streamer", obj = "/" },
	["streamer"] = { bus = "streamer", obj = "/DefaultApp" },
	["media_users"] = { bus = "streamer", obj = "/MediaUsers" },
	["audioEngine"] = { bus = "streamer", obj = "/AudioEngine" },
	["avio"] = { bus="codec", obj = "/AVIO" },
	["osd"] = { bus="codec", obj = "/OSD" },
	["onvif"] = { bus="onvif", obj="/" },
	["ptz"] = { bus="onvif", obj="/PTZ" },
	["hostapm"] = { bus="nm_daemon", obj="/raw/HOSTAP" },
	["record_m"] = { bus="record", obj="/" },
	["webConfig"] = { bus="web_daemon", obj="/" }
}
---------------------------------------------------

local function ParseParamTarget(idr)
	if idr:match("^_ENV%.") then
		local targ = idr:match("^_ENV%.(.-)%.")
		if not targ or not getSetTargets[targ] then
			return false,nil
		else
			return true,targ
		end
	else
		local targ = idr:match("^(.-)%.")
		if not targ or not getSetTargets[targ] then
			return false,nil
		else
			return false,targ
		end
	end
end

---------------------------------------------------

local function GroupParams(params)
	local group = {}
	local k,v
	for k,v in pairs(params) do
		local isEnv, targ = ParseParamTarget(k)
		if targ then
			if not group[targ] then
				group[targ] = { env={}, env_cc = 0, conf={}, conf_cc = 0 }
			end

			if isEnv then
				group[targ].env[k] = v
				group[targ].env_cc = group[targ].env_cc + 1
			else
				group[targ].conf[k] = v
				group[targ].conf_cc = group[targ].conf_cc + 1
			end
		end
	end
	return group
end

-----------------------------------------------

function _M.Get(params)
	local group = GroupParams(params)
	local k,v
	local cc = 0

	for k,v in pairs(group) do
		if getSetTargets[k].isLocal then --Local get
			if v.conf_cc > 0 then
				getset.GetValues( k, v.conf )
				local kk, vv
				for kk, vv in pairs(v.conf) do
					params[kk] = vv
					cc = cc + 1
				end
			end
			
			if v.env_cc > 0 then
				getset.GetValues( "_ENV", v.env )
				local kk, vv
				for kk, vv in pairs(v.conf) do
					params[kk] = vv
					cc = cc + 1
				end
			end
		else --Remote GET
			local proxy = GetCoAFProxy( getSetTargets[k].obj, getSetTargets[k].bus )
			if proxy and v.conf_cc > 0 then
				local r,x,res = proxy:Get( v.conf )
				if r and x and x >= 0 and res then
					local kk,vv
					for kk,vv in pairs(res) do
						params[kk] = vv
						cc = cc + 1
					end
				end
			end

			if proxy and v.env_cc > 0 then
				local r,x,res = proxy:GetEnv( v.env )
				if r and x and x >= 0 and res then
					local kk,vv
					for kk,vv in pairs(res) do
						params[kk] = vv
						cc = cc + 1
					end
				end
			end

			if proxy then
				proxy:Destroy()
			end
		end
	end

	return cc, params
end

-----------------------------------------------

function _M.Set(params)
	local group = GroupParams(params)
	local k,v
	local result = false
	for k,v in pairs(group) do
		if getSetTargets[k].isLocal then --Local SET
			if v.conf_cc > 0 then
				getset.SetConfigureValues( k, v.conf )
				result = true
			end
			
			if v.env_cc > 0 then
				getset.SetValues( "_ENV", v.env )
				result = true
			end
		else --Remote SET
			local proxy = GetCoAFProxy( getSetTargets[k].obj, getSetTargets[k].bus )

			if proxy and v.conf_cc > 0 then
				proxy:Set( v.conf )
				result = true
			end

			if proxy and v.env_cc > 0 then
				proxy:SetEnv( v.env )
				result = true
			end

			if proxy then
				proxy:Destroy()
			end
		end
	end

	return result
end

-----------------------------------------------

function _M.Init(opts)
	if type(opts) ~= "table" then
		opts = {
			local_get_set = nil
		}
	end

	NETWORK_MANAGER.Init()
	_M.NETWORK_MANAGER = NETWORK_MANAGER

	if opts.local_get_set then
		if type(opts.local_get_set) == "string" then
			if getSetTargets[opts.local_get_set] then
				getSetTargets[opts.local_get_set].isLocal = true
			end
		elseif type(opts.local_get_set) == "table" then
			local i,v
			for i,v in pairs(opts.local_get_set) do
				if type(v) == "string" and getSetTargets[v] then
					getSetTargets[v].isLocal = true
				elseif type(i) == "string" and getSetTargets[i] then
					getSetTargets[i].isLocal = true
				end
			end
		end
	end
end

-----------------------------------------
--- Modules reset
-----------------------------------------

function _M.Reset()
	if reset then
		print( "*** Quick Reset the system" )
		reset.Reset( clientSession )
	end
	return true
end

function _M.Reboot()
	if reset then
		print( "*** Reboot the device" )
		--reset.Reboot()
		_G.REBOOT_REQUIRE = true
	end
end

-----------------------------------------
--- Get signal status
-----------------------------------------

function _M.GetSignalStatus()
	local vfeProxy = GetCoAFProxy( "/AVIO/00/VFE", "codec" )
	if not vfeProxy then
		print( "*** ERROR: Could not get VFE proxy when GetSignalStatus()" )
		return {}
	end

	local result = {}
	local r, x, rtable = vfeProxy:Control( "detect-source", {source="any"} )
	vfeProxy:Destroy()
	if r and x and x >= 0 then
		result.video = rtable
		if rtable["video.locked"] == 1 and rtable["video.freerun"] and rtable["video.freerun"] == 1 then	--Processing the special case: freerun.
			rtable["video.locked"] = 2
		end
	else
		print( "***vfe:detect-source WRONG", x, rtable )
	end

	local afeProxy = GetCoAFProxy( "/AVIO/00/AFE", "codec" )
	if not afeProxy then
		print( "*** ERROR: Could not get AFE proxy when GetSignalStatus()" )
		return result
	end

	r, x, rtable = afeProxy:Control( "detect-source", {source="any"} )
	afeProxy:Destroy()
	if r and x and x >= 0 then
		result.audio = rtable
	else
		print( "***afe:detect-source WRONG", x, rtable )
	end

	return result
end

-----------------------------------------
--- Get Stream Status
-----------------------------------------

function _M.VGASpecialCtrl(flag, special, params)
	if not flag or not special or type(params) ~= "table" then
		return nil
	end
	
	local vfeProxy = GetCoAFProxy( "/AVIO/00/VFE", "codec" )
	if not vfeProxy then
		print( "*** ERROR: Could not get VFE proxy when GetSignalStatus()" )
		return {}
	end

	params["special"] = special
	params["source"] = "VGA"
	
	if flag == "set" or flag == "SET" then
		local r,x,rtable = vfeProxy:Control( "config-source", params )
		vfeProxy:Destroy()
		--print( r,x,rtable)
		if not r or not x then
			return nil
		else
			return rtable
		end
	else
		local r,x,rtable = vfeProxy:Control( "detect-source", params )
		vfeProxy:Destroy()
		if not r or not x then
			return nil
		else
			return rtable
		end
	end
end

-----------------------------------------
--- Update encoding stream
-----------------------------------------
function _M.UpdateEncoding(idr, target)
	local vidEncProxy = GetCoAFProxy( "/Codec/Encoders/" .. idr, "codec" )
	if vidEncProxy then
		local r, x = vidEncProxy:UpdateEncoding(target or "all")
		vidEncProxy:Destroy()
		return x
	else
		return false
	end
end

-----------------------------------------
--- LED User controls
-----------------------------------------

function _M.SetLEDUserControl(en)
	local ioledProxy = GetCoAFProxy( "/", "ioled" )
	if not ioledProxy then
		return false
	end

	local ctrl = "disable_uctrl"

	if en and en ~= 0 then
		ctrl = "enable_uctrl"
	end

	local r,x = ioledProxy:Control( ctrl, {} )
	ioledProxy:Destroy()
	if r and x then
		return true
	else
		return false
	end
end

function _M.LEDControl(id,status)
	local ioledProxy = GetCoAFProxy( "/", "ioled" )
	if not ioledProxy then
		return false
	end

	local r,x = ioledProxy:Control( "led_uctrl", { id=id or "", status = status or "OFF" })
	ioledProxy:Destroy()
	if r and x then
		return true
	else
		return false
	end
end

function _M.PTZControl(action,args)
	local ptzProxy = GetCoAFProxy( "/PTZ", "onvif" )
	if not ptzProxy then
		return false
	end

	local r,x = ptzProxy:Control( action, args)
	ptzProxy:Destroy()
	if r and x then
		return true
	else
		return false
	end
end
 
function _M.IRControl(action,args)
	local irProxy = GetCoAFProxy( "/", "ir" )
	if not irProxy then
		return false
	end

	local r,x,rtable = irProxy:Control( action, args)
	irProxy:Destroy()
	if r and x then
		return x,rtable
	else
		return false
	end
end

-----------------------------------------
--- Get Record Status
-----------------------------------------

function _M.GetAllDisks()
	local diskProxy = GetCoAFProxy( "/", "record" )
	if not diskProxy then
		return nil, nil
	end
	local r,st,info = diskProxy:DiskList()
	if not r or not st or type(info) ~= "table" then
		return nil, nil
	end

	local rslt = {}
	local cc = 0
	local k,v
	for k,v in pairs(info) do
		local detail
		r,st,detail = diskProxy:DiskDetail(k)
		if r and st and type(detail) == "table" then
			rslt[k] = detail
			cc = cc + 1
		end
	end
	diskProxy:Destroy()

	return cc, rslt
end

function _M.DiskRecordList(dev,filter)
	local diskProxy = GetCoAFProxy( "/", "record" )
	if not diskProxy or not dev or dev == "" then
		return nil, nil
	end
	local r,st,info = diskProxy:DiskRecordList(dev, filter or "")
	diskProxy:Destroy()
	if not r or not st or type(info) ~= "table" then
		return nil, nil
	end

	local rslt = {}
	local k,v
	for k,v in pairs(info) do
		if v.FILES then
			local files = v.FILES .. "\n"
			v.FILES = {}
			local f,s
			for f,s in files:gmatch("(.-)=(.-)\n") do
				table.insert( v.FILES, { FILENAME=f, SIZE=s } )
			end
			table.sort( v.FILES, function(a,b)
				return a.FILENAME < b.FILENAME
			end)
			table.insert( rslt, v)
		end
	end

	table.sort( rslt, function(a,b)
		return (a.START_TIME or 0) > (b.START_TIME or 0)
	end)
	return #rslt, rslt
end

function _M.GetRecordStatus( idr )
	local idrNum = tonumber(idr) or 0
	local recordProxy = GetCoAFProxy( "/", "record" )
	if not recordProxy then
		return nil, nil
	end
	local r,st,info = recordProxy:GetStatus( idrNum == 1 and "sub" or "main" )
	recordProxy:Destroy()
	if not r or type(info) ~= "table" then
		return nil, nil
	else
		if info.FILES then
			local files = info.FILES .. "\n"
			info.FILES = {}
			local f,s
			for f,s in files:gmatch("(.-)=(.-)\n") do
				table.insert( info.FILES, { FILENAME=f, SIZE=s } )
			end
			table.sort( info.FILES, function(a,b)
				return a.FILENAME < b.FILENAME
			end)
		end
		return st, info
	end
end

-----------------------------------------
--- Get Stream Status
-----------------------------------------

function _M.GetStreamStatus( idr )
	local srcProxy = GetCoAFProxy( "/DefaultApp/" .. idr .. "/Source", "streamer" )
	local vidEncProxy = GetCoAFProxy( "/Codec/Encoders/" .. idr, "codec" )

	local result = {}

	local idx = tonumber(idr)
	-- TODO: Audio
	local params = {
		["encoder.stream[" .. idx .. "].enabled"] = 0,
		["encoder.stream[" .. idx .. "].video.profile"] = "baseline",
		["encoder.stream[" .. idx .. "].video.scale"] = "default",
		["encoder.stream[" .. idx .. "].video.codec"] = "NONE",
		["streamer.stream[" .. idx .. "].audioChannel"] = -1
	}
	
	local gotAudio = false
	local r
	r,params = _M.Get(params)
	
	if r and r >=0 and params then
		result["stream.enabled"] = params["encoder.stream[" .. idx .. "].enabled"]
		result["video.codec"] = params["encoder.stream[" .. idx .. "].video.codec"]
		result["video.scale"] = params["encoder.stream[" .. idx .. "].video.scale"]
		result["video.profile"] = params["encoder.stream[" .. idx .. "].video.profile"]
		local aChannel = params[ "streamer.stream[" .. idx .. "].audioChannel" ] or -1
		result["audio.channel"] = aChannel
		if aChannel >= 0 then
			params = {
				["audioEngine.encoders[" .. aChannel .. "].encoder.sampling"] = 48000,
				["audioEngine.encoders[" .. aChannel .. "].encoder.channels"] = 2,
				["audioEngine.encoders[" .. aChannel .. "].encoder.codec"] = "AAC",
				["audioEngine.encoders[" .. aChannel .. "].encoder.bitrate"] = 64000
			}
			r,params = _M.Get(params)
			if r and r >=0 and params then
				gotAudio = true
				result["audio.codec"] = params["audioEngine.encoders[" .. aChannel .. "].encoder.codec"] or "NONE"
				result["audio.sampling"] = params["audioEngine.encoders[" .. aChannel .. "].encoder.sampling"]  or 0
				result["audio.channels"] = params["audioEngine.encoders[" .. aChannel .. "].encoder.channels"] or 0
				result["audio.bitrate"] = params["audioEngine.encoders[" .. aChannel .. "].encoder.bitrate"] or 0
			end
		end
		
		if not gotAudio then
			--TODO:
			result["audio.codec"] = "NONE"
			result["audio.sampling"] = 0
			result["audio.channels"] = 0
			result["audio.bitrate"] = 0
		end
		
	else
		print( "*** Fail to get encoder settings of '" .. idr .. "'" )
	end

	if vidEncProxy then
		local r,x,params = vidEncProxy:GetVideoProperties()
		vidEncProxy:Destroy()
		if r and x and x >= 0 and params then
			result["video.width"] = params["video.width"]
			result["video.height"] = params["video.height"]
		else
			print( "*** Could not get video encoding informations of '" .. idr .. "'" )
		end
	else
		print( "*** Could not get encoder control proxy of '" .. idr .. "'" )
	end

	if srcProxy then
		local r,x,info = srcProxy:GetStatus()
		srcProxy:Destroy()
		if r and x and x >=0 and info then
			result["video.bitrate"] = info["video.bitrate"]
			result["video.frameRate"] = info["video.frameRate"]
		else
			print( "*** Stream source of '" .. idr .. "': Fail to get status" )
		end
	else
		print( "*** Could not get source stream proxy of '" .. idr .. "'" )
	end
	return result
end

-----------------------------------------

function _M.GetEncoderStreamStatus( idr )
	
	local vidEncProxy = GetCoAFProxy( "/Codec/Encoders/" .. idr, codec )

	local result = {}

	local idx = tonumber(idr)
	-- TODO: Audio
	local params = {
		["encoder.stream[" .. idx .. "].enabled"] = 0,
		["encoder.stream[" .. idx .. "].video.profile"] = "baseline",
		["encoder.stream[" .. idx .. "].video.scale"] = "default",
		["encoder.stream[" .. idx .. "].video.codec"] = "NONE",
		["encoder.stream[" .. idx .. "].video.bitrate"] = 0,
		["streamer.stream[" .. idx .. "].audioChannel"] = -1
	}
	
	local gotAudio = false
	local r
	r,params = _M.Get(params)
	
	if r and r >=0 and params then
		result["stream.enabled"] = tonumber(params["encoder.stream[" .. idx .. "].enabled"])
		result["video.codec"] = params["encoder.stream[" .. idx .. "].video.codec"]
		result["video.scale"] = params["encoder.stream[" .. idx .. "].video.scale"]
		result["video.profile"] = params["encoder.stream[" .. idx .. "].video.profile"]
		result["video.bitrate"] = params["encoder.stream[" .. idx .. "].video.bitrate"]
		
		local aChannel = params[ "streamer.stream[" .. idx .. "].audioChannel" ] or -1
		result["audio.channel"] = aChannel
		if aChannel >= 0 and _G.audioEngine then
			gotAudio = true
			result["audio.codec"] = _G.audioEngine.encoders[aChannel].encoder.codec or "NONE"
			result["audio.sampling"] = _G.audioEngine.encoders[aChannel].encoder.sampling or 0
			result["audio.channels"] = _G.audioEngine.encoders[aChannel].encoder.channels or 0
			result["audio.bitrate"] = _G.audioEngine.encoders[aChannel].encoder.bitrate or 0
		end
		
		if not gotAudio then
			--TODO:
			result["audio.codec"] = "NONE"
			result["audio.sampling"] = 0
			result["audio.channels"] = 0
			result["audio.bitrate"] = 0
		end
	else
		print( "*** Fail to get encoder settings of '" .. idr .. "'" )
		result["stream.enabled"] = 0
		result["video.codec"] = "H264"
		result["video.scale"] = "default"
		result["video.profile"] = "baseline"
		result["audio.channel"] = -1
		result["audio.codec"] = "NONE"
		result["audio.sampling"] = 0
		result["audio.channels"] = 0
		result["audio.bitrate"] = 0
	end

	if vidEncProxy then
		local r,x,params = vidEncProxy:GetVideoProperties()
		vidEncProxy:Destroy()
		if r and x and x >= 0 and params then
			result["video.width"] = params["video.width"]
			result["video.height"] = params["video.height"]
			result["video.framerate"] = params["video.framerate"]
		else
			print( "*** Could not get video encoding informations of '" .. idr .. "'" )
			result["video.width"] = 0
			result["video.height"] = 0
			result["video.framerate"] = 0
		end
	else
		print( "*** Could not get encoder control proxy of '" .. idr .. "'" )
		result["video.width"] = 0
		result["video.height"] = 0
		result["video.framerate"] = 0
	end

	return result
end

-----------------------------------------
--- Stream snapshot
-----------------------------------------
function _M.SnapShot( idr, path, file, ext )
	if not idr then
		print( "*** Invalid IDR when snap stream" )
		return false
	end

	local moduleProxy = GetCoAFProxy( "/DefaultApp/" .. idr, "streamer" )
	if not moduleProxy then
		print( "*** Could not get streaming service module proxy of stream '" .. idr .. "'" )
		return false
	end

	local r,cc,msg = moduleProxy:Control( "snap", {
		["path"] = path or "/tmp",
		["file"] = file or "WEB-SNAP",
		["suffix"] = ext or ".jpg",
		["count"] = 1
	} )
	moduleProxy:Destroy()

	if not r or not cc or cc < 0 then
		return false
	else
		return true
	end
end

-----------------------------------------
--- Get stream service status
-----------------------------------------
function _M.GetStreamServiceStatus( idr, service )
	if not idr or not service then
		print( "*** Invalid IDR/Service name when getting stream service" )
		return {}
	end

	local svcProxy = GetCoAFProxy( "/DefaultApp/" .. idr .. "/" .. service, "streamer" )
	if not svcProxy then
		print( "*** Could not get streaming service proxy of stream '" .. idr .. "', service='" .. service .. "'" )
		return {}
	end

	local r,x,status = svcProxy:GetStatus()
	if r and x and x >=0 and status then
		status["enabled"],status["serviceUri"],status["serviceType"] = svcProxy._P("Enabled","ServiceUri", "ServiceType")
		svcProxy:Destroy()
		return status
	else
		svcProxy:Destroy()
		print( "*** Could not get streaming service status of stream '" .. idr .. "', service='" .. service .. "'" )
		return {}
	end
end

-----------------------------------------
--- Control stream service enable
-----------------------------------------
function _M.ControlStreamServiceEnable( idr, service, en )
	if not idr or not service then
		print( "*** Invalid IDR/Service name when getting stream service" )
		return false
	end

	local svcProxy = GetCoAFProxy( "/DefaultApp/" .. idr .. "/" .. service, "streamer" )
	
	if not svcProxy then
		print( "*** Could not get streaming service proxy of stream '" .. idr .. "', service='" .. service .. "'" )
		return false
	end

	local r,msg = svcProxy:Enable(en)
	svcProxy:Destroy()
	if r then
		return true
	else
		print( "*** Could not control streaming service enable/disable of stream '" .. idr .. "', service='" .. service .. "'" )
		print(msg)
		return false
	end	
end

-----------------------------------------------

function _M.LoadPreset( id, loc )
	return presetM.LoadPreset(id, loc)
end

-----------------------------------------------

function _M.UpdateStreamIdr( streamId )
	local proxy = GetCoAFProxy( "Codec/Encoders/" .. streamId, "codec" )	
	if proxy then
		proxy:UpdateEncoding("video")
	end
end

-----------------------------------------------

function _M.GetIfDevices( filter )
	return NETWORK_MANAGER.GetDevicesBy( filter )
end

-----------------------------------------------

function _M.GetIfDeviceInfo( obj, filter )
	return NETWORK_MANAGER.GetDeviceInfo( obj, filter )
end

-----------------------------------------------
function _M.UpdateIfConfig( obj )
	NETWORK_MANAGER.ConnectDevice( obj )
end

-----------------------------------------------

function _M.SetWireIfConfig( args )
	if not args or type(args) ~= "table" then
		return false
	end

	if not args.obj then
		return false
	end

	local devInfo = NETWORK_MANAGER.GetDeviceInfo( args.obj, {WIRE = true} )	--Get WIRE informations, so I can get MAC.
	if not devInfo then
		return false
	end

	local devName = devInfo.Name

	local connSettings = {
		["connection"] = {},
		["802-3-ethernet"] = {},
		["ipv4"] = {}
	}

	connSettings["connection"]["type"] = "802-3-ethernet"
	connSettings["connection"]["autoconnect"] = true	--Force auto connection.

	if args.uuid and args.uuid ~= "" then
		connSettings["connection"]["uuid"] = args.uuid
		local detail = NETWORK_MANAGER.ListConnectionAndGetDetails( {uuid=args.uuid} )
		if detail then
			local k,v
			local found
			for k,v in pairs(detail) do
				found = v
				break
			end
			
			if found then
				connSettings["connection"]["id"] = found["connection"]["id"]
			end
		end
	end

	if not connSettings["connection"]["id"] then
		connSettings["connection"]["id"] = devName
	end
	
	if args.cloneMac then
		if args.cloneMac ~= "" then
			connSettings["802-3-ethernet"]["cloned-mac-address"] = nu.MACToArray(args.cloneMac)
		else
			connSettings["802-3-ethernet"]["cloned-mac-address"] = nu.MACToArray(devInfo.WIRE.PermHwAddress)
		end
	end

	connSettings["connection"]["interface-name"] = devName

	if args.dynamic and args.dynamic == "y" then	--DHCP
		connSettings["ipv4"]["method"] = "auto"
	else
		connSettings["ipv4"]["method"] = "manual"
		if args.ip and args.ip ~= "" then
			connSettings["ipv4"]["address-data"] = {{address=args.ip, prefix=nu.ip4MaskToPrefix(args.mask)}}
		end

		connSettings["ipv4"]["gateway"] = args.gw or ""

		if args.dns1 and args.dns1 == "" then
			args.dns1 = nil
		end

		if args.dns2 and args.dns2 == "" then
			args.dns2 = nil
		end

		connSettings["ipv4"]["dns"] = { args.dns1 or nil, args.dns2 or nil }
		if #connSettings["ipv4"]["dns"] < 1 then
			connSettings["ipv4"]["dns"] = nil
		end
	end

	local o = NETWORK_MANAGER.ChangeConnectionDetails( args.specConn, connSettings )
	if o then
		return true
	else
		return false
	end
end

-----------------------------------------------

function _M.CheckIfConfig(obj, type)
	local specConn
	type = type or "WIRE"
	local devInfo = NETWORK_MANAGER.GetDeviceInfo( obj, {[type] = true} )	--Get spec type informations, so I can get MAC.
	if not devInfo then
		return nil
	end

	if devInfo.ActiveConnection and devInfo.ActiveConnection ~= "" and devInfo.ActiveConnection ~= "/" then
		local ac = NETWORK_MANAGER.GetActivateConnectionDetails(devInfo.ActiveConnection)
		if ac then
			specConn = ac[devInfo.ActiveConnection].Connection
		end
	end

	if not specConn and type == "WIRE" then
		specConn = devInfo.AvailableConnections[1]
		if not specConn then
			local allConns = NETWORK_MANAGER.ListConnectionAndGetDetails( { interface=devInfo.Name } )
			if allConns then
				local k,v
				for k,v in pairs(allConns) do
					specConn = k	--Get first one.
					break
				end
			end
		end
	end

	local devMac = devInfo[type].PermHwAddress or ""	--Identify the device.
	local dhcp = true
	local ip4addr = ""
	local ip4mask = ""
	local ip4gateway = ""
	local ip4dns1 = ""
	local ip4dns2 = ""
	local prefix = 24
	local uuid = ""
	local cloneMac = ""

	if specConn then
		local connInfo = NETWORK_MANAGER.GetConnectionDetails( specConn )
		if not connInfo or not connInfo[specConn] then
			return nil
		end

		connInfo = connInfo[specConn]

		uuid = connInfo["connection"]["uuid"]
		dhcp = connInfo["ipv4"]["method"] and connInfo["ipv4"]["method"] == "auto" or false
		if type == "WIRE" then
			cloneMac = connInfo["802-3-ethernet"]["cloned-mac-address"] or connInfo["802-3-ethernet"]["mac-address"]
		else
			cloneMac = ""
		end

		if cloneMac and #cloneMac >= 6 then
			cloneMac = string.format( "%02X:%02X:%02X:%02X:%02X:%02X", cloneMac[1],cloneMac[2],cloneMac[3],cloneMac[4],cloneMac[5],cloneMac[6] )
		else
			cloneMac = ""
		end

		if connInfo["ipv4"]["address-data"] and #connInfo["ipv4"]["address-data"] > 0 then
			ip4addr = connInfo["ipv4"]["address-data"][1]["address"]
			local pfx = connInfo["ipv4"]["address-data"][1]["prefix"]
			prefix = pfx
			ip4mask = nu.ip4PrefixToMask(pfx)
			ip4gateway = connInfo["ipv4"]["gateway"]

			if connInfo["ipv4"]["dns"] and #connInfo["ipv4"]["dns"] > 0 then
				ip4dns1 = connInfo["ipv4"]["dns"][1] or ""
				ip4dns2 = connInfo["ipv4"]["dns"][2] or ""
			end
		end
	end
	
	return {
		obj = obj,
		specConn = specConn,
		devMac = devMac,
		dhcp = dhcp,
		ip4addr = ip4addr,
		ip4mask = ip4mask,
		prefix = prefix,
		ip4gateway = ip4gateway,
		ip4dns1 = ip4dns1,
		ip4dns2 = ip4dns2,
		uuid = uuid,
		cloneMac = cloneMac
	}
end

-----------------------------------------------
--[[ JUNS args:
*****视频参数
-vb 视频码率
-s 视频分辨率
-r 视频帧率
-cvbr 码率控制方式，参数：cbr,vbr
-profile 档次，参数：baseline,main,high
-gop I帧间隔（秒）

*****音频参数
-ab 音频码率
-ar 音频采集率
-ac 音频声道，参数：0：左声道，1：右声道，2：立体声道
-as 音频信号源，参数：sdi, linein
-av 音频增溢

如：
-vb 500k -s 640x360 -r 15 -cvbr cbr -profile baseline -gop 2 -ab 48k -ar 44100 -ac 2 -as sdi -av 0
]]
local JUNSArgsMap = {
	["-vb"] = { path="encoder.stream[%d].video.bitrate", vArgs = { "id" }, bitrate = true },
	["-s"] = { path="encoder.stream[%d].video.scale", vArgs = { "id" }, resolution = true },
	["-r"] = { path="encoder.stream[%d].video.framerate", vArgs = { "id" } },
	["-cvbr"] = { path="encoder.stream[%d].video.br_ctrl", vArgs = { "id" } },
	["-profile"] = { path="encoder.stream[%d].video.profile", vArgs = { "id" } },
	["-gop"] = { path="encoder.stream[%d].video.gop_size", vArgs = { "id" }, gop = true },
	["-ab"] = { path="audioEngine.encoders[%d].encoder.bitrate", vArgs = { "ae" }, bitrate = true },
	["-ar"] = { path="audioEngine.encoders[%d].encoder.sampling", vArgs = { "ae" }, sampling = true },
	["-ac"] = { path="audioEngine.encoders[%d].encoder.channels", vArgs = { "ae" } },
	["-as"] = { path="avio.audio.source" },
	["-av"] = { path="avio.audio.line_gain" },
}	

function _M.SetJUNSArgs( streamId, args )
	if not streamId or type(args) ~= "string" then
		return false
	end
	
	print( "Setting JUNS-ARGs: ", args )
	
	streamId = tonumber(streamId)
	if streamId < 0 or streamId > 2 then
		print( "ERROR: Set JUNS arguments but invalid streamId passed (" .. streamId .. ")" )
		return false
	end
	
	local streamAE = _G.streamer.stream[streamId].audioChannel or -1
	
	local parsed_args = {}
	local k,v
	local argsCC = 0
	for k,v in args:gmatch( "(%-%a-)%s+(%g+)" ) do
		local item = JUNSArgsMap[k]
		if not item or not v then
			print( "WARNING: JUNS-ARG (" .. (k or "?") .. "=" .. (v or "") .. ") is not support or invalid!" )
		else
			print( "JUNS-ARG: " .. k .. "=" .. v )
			local path = item.path
			--Something variant arguments.
			local vArgs = {}
			if item.vArgs then
				local i
				for i=1, #item.vArgs do
					if item.vArgs[i] == "id" then
						table.insert(vArgs, streamId )
					elseif item.vArgs[i] == "ae" then
						table.insert(vArgs, streamAE )
					end
				end
			end
			if #vArgs > 0 then
				path = path:format( table.unpack(vArgs) )
			end
			
			if item.bitrate then --Parse bitrate
				local unit = v:match( "%d+([kKmM])" )
				if unit then
					unit = unit:lower()
					v = v:match( "(%d+)" )
					if not v then
						v = 0
					else
						v = tonumber(v) * ( (unit == 'k') and 1000 or 1000000 )
					end
				else
					v = tonumber(v)
				end
			elseif item.resolution then --Parse resolution
				local x, y = v:match( "(%d+)[xX](%d+)" )
				if not x or not y then
					v = v:lower()
					if v == "1080p" or v == "1080i" then
						v = "1920x1080"
						x = 1920
						y = 1080
					elseif v == "720p" then
						v = "1280x720"
						x = 1280
						y = 720
					elseif v == "d1" then
						v = "720x480"
						x = 720
						y = 480
					elseif v == "cif" then
						v = "352x288"
						x = 352
						y = 288
					else
						v = "default"
						x = 1920 --I'dont know, so assume a big value.
						y = 1080
					end
					
					if streamId == 0 and ( x < 720 or y < 480 ) then --The main stream can only support > 720x480
						v = "720x480"
					end
				else
					x = tonumber(x)
					y = tonumber(y)
					if streamId == 0 and ( x < 720 or y < 480 ) then --The main stream can only support > 720x480
						v = "720x480"
					else
						v = x .. "x" .. y
					end
				end
			elseif item.gop then
				--TODO: I'm lazy to detect framerate so assume it be 25. Hmm...
				v = tonumber(v)
				if v <= 0 then
					v = 25
				else
					v = 25*v
				end
			elseif item.sampling then
				--TODO: Now only support 48000, so ignore this.
				v = 48000
			end
			print( "JUNS-ARG: ==> " .. path .. " = " .. v )
			parsed_args[path] = v
			argsCC = argsCC + 1
		end
	end --For each arguments checking.
	
	if argsCC <= 0 then
		print( "WARNING: No JUNS-ARGs parsed, ignored setting." )
		return true
	else
		_M.Set( parsed_args )
		return true
	end
end

-----------------------------------------------

return _M

