--[==[

The Co-Application Framework (CoAF) thread (That's called 'CoThread').

The CoAF modules *MUST* depend on 'lua-ev' module.

Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.

]==]

local _dbg = require "coaf.debug"
local _flag, coaf = pcall(require, "coaf.core")  --try require it, but not neccesary.
if not _flag then
	coaf = nil
end

---------------------------------------------------------

local thread_meta = {}

function thread_meta:Start(...)
	if not coaf then
		return self:StartNow(...) --A bug fixed here
	end

	-- Only if 'CoAF' core is required.

	if self.started then
		return
	end

	local args = {...}
	local coro = self.coro
	if self.dont_pass_self then
		coaf.ScheduleDelayedTask(function()
			--If terminated before here, return.
			if self.terminated then
				return
			end
			self.__actual_started = true
			local r, st = coroutine.resume(coro, table.unpack(args))
			if not r then
				_dbg.Error("CoAF", "Thread error:", st)
			end
		end)
	else
		coaf.ScheduleDelayedTask(function()
			--If terminated before here, return.
			if self.terminated then
				return
			end
			self.__actual_started = true
			local r, st = coroutine.resume(coro, self, table.unpack(args))
			if not r then
				_dbg.Error("CoAF", "Thread error:", st)
			end
		end)
	end
	self.started = true
end

function thread_meta:StartNow(...)
	if self.started then return end
	self.started = true
	if self.dont_pass_self then
		self.__actual_started = true
		local r, st = coroutine.resume( self.coro, ... )
		if not r then
			_dbg.Error("CoAF", "Thread error:", st)
		end
		return r
	else
		self.__actual_started = true
		local r, st = coroutine.resume( self.coro, self, ... )
		if not r then
			_dbg.Error("CoAF", "Thread error:", st)
		end
		return r
	end
end

function thread_meta:Terminate()
	self.terminated = true
	local status = coroutine.status( self.coro )
	if self.__actual_started and status == "suspended" then -- Try 'KILL' (#CoAF-KILL#) the pending task.
		local r, st = coroutine.resume(self.coro, nil, "#CoAF-KILL#")
		if not r then
			_dbg.Error("CoAF", "Terminate thread error:", st)
		end
	end
	self.started = false
end

function thread_meta:Kill(sig)
	local status = coroutine.status( self.coro )
	if self.__actual_started and status == "suspended" then -- Try KILL(signal) the pending task.
		local r, st = coroutine.resume(self.coro, false, "#CoAF-KILL#", sig )
		if not r then
			_dbg.Error("CoAF", "Kill thread error:", st)
		end
	end
end

function thread_meta:Status()
	local status = coroutine.status(self.coro)
	if status == "running" and self.terminated then
		status = "killing"
	end
	return status
end

---------------------------------------------------------

local _M = {}

-- New a thread, to call function 'func'. if 'dont_pass_self' is
-- true, the first argument of 'func' is not thread object itself;
-- otherwise, the first argument of 'func' is thread object.

function _M.New(func, dont_pass_self)
	local coro = coroutine.create(func)
	if not coro then
		_dbg.Error("CoAF", "Fail to create coroutine for thread")
		return nil
	end
	local threadObj = {
		coro = coro,
		dont_pass_self = dont_pass_self,
		terminated = false,
		started = false,

		__actual_started = false
	}
	setmetatable(threadObj, {__index = thread_meta})
	return threadObj
end

-- A new feature, to support thread-based timer.

function _M.NewTimer(func, delay, period)
	if not coaf then
		_dbg.Error( "CoAF", "The thread-based timer MUST use CoAF framework." )
		return nil
	end

	if type(func) ~= "function" then
		_dbg.Error( "CoAF", "thread.NewTimer(...) requires a function for argument 1" )
		return nil
	end

	local posix_time = require "posix.time"
	local nextActiveTime = posix_time.clock_gettime( posix_time.CLOCK_MONOTONIC )
	nextActiveTime.tv_sec = nextActiveTime.tv_sec + math.floor(delay)
	nextActiveTime.tv_nsec = nextActiveTime.tv_nsec + math.floor( (delay-math.floor(delay)) * 1000000000 )
	if nextActiveTime.tv_nsec >= 1000000000 then
		nextActiveTime.tv_sec = nextActiveTime.tv_sec + 1
		nextActiveTime.tv_nsec = nextActiveTime.tv_nsec - 1000000000
	end


	local tobj = _M.New( function(thrObj)
		local now = posix_time.clock_gettime( posix_time.CLOCK_MONOTONIC )
		local diff = nextActiveTime.tv_sec - now.tv_sec
		diff = diff + (nextActiveTime.tv_nsec - now.tv_nsec)/1000000000

		if not thrObj.terminated and diff > 0 then
			coaf.Sleep( diff )
		end

		while not thrObj.terminated do

			func(thrObj)

			if type(period) ~= "number" or period <= 0 then
				break
			end

			nextActiveTime.tv_sec = nextActiveTime.tv_sec + math.floor(period)
			nextActiveTime.tv_nsec = nextActiveTime.tv_nsec + math.floor( (period - math.floor(period)) * 1000000000 )
			if nextActiveTime.tv_nsec >= 1000000000 then
				nextActiveTime.tv_sec = nextActiveTime.tv_sec + 1
				nextActiveTime.tv_nsec = nextActiveTime.tv_nsec - 1000000000
			end
			now = posix_time.clock_gettime( posix_time.CLOCK_MONOTONIC )

			diff = nextActiveTime.tv_sec - now.tv_sec
			diff = diff + (nextActiveTime.tv_nsec - now.tv_nsec)/1000000000

			if diff > 0 then
				coaf.Sleep(diff)
			else
				nextActiveTime = now
				--To yield task.
				coaf.ScheduleDelayedTask( function()
					local status = coroutine.status( thrObj.coro )
					if status == "suspended" then 
						local r, st = coroutine.resume(thrObj.coro,true)
						if not r then
							_dbg.Error("CoAF", "Thread error:", st)
						end
					end
				end, 0 )

				local r,flag = coroutine.yield(coro)
				if not r and flag == "#CoAF-KILL#" then
					break
				end
			end
		end
		return true
	end)
	--tobj:Start()
	--For compatible with coaf.NewTimer(...)
	tobj.start = tobj.Start
	tobj.stop = tobj.Terminate
	tobj.destory = tobj.Terminate
	tobj.again = function()
		_dbg.Warn( "CoAF", "Thread-based timer does not support 'again' operation" )
	end
	return tobj
end


return _M


