local coaf = require "coaf.core"
local _dbg = require "coaf.debug"
local st, posix_unistd = pcall(require,"posix.unistd")
local pretty = require "pl.pretty"
local _M = {}

local action_meta = {}

function action_meta:__call(svcConfig, actionFlag, actionArg)
	if actionFlag and actionFlag ~= "start" then
		if actionFlag == "stop" then
			if self.doStopAction then
				return self:doStopAction(svcConfig)
			else
				return false
			end
		elseif actionFlag == "signal" then
			if self.doSignalAction then
				return self:doSignalAction(svcConfig, actionArg)
			else
				return false
			end
		else
			return false
		end
	elseif self.doAction then
		return self:doAction(svcConfig)
	else
		return false
	end
end

function action_meta:__tostring()
	if self.type == "EXEC" or self.type == "EXECP" then
		local rslt = {self.type, "( '"}
		table.insert(rslt, self.cmd or "")
		table.insert(rslt, "'")
		if self.args then
			table.insert(rslt,", ")
			table.insert(rslt, pretty.write(self.args,''))
		end
		table.insert(rslt, " )")
		
		return table.concat(rslt)
	elseif self.type == "ACCEPT_PROXY" then
		local rslt = {self.type, "("}
		if not self.addressPattern then
			table.insert(rslt, "nil")
		else
			table.insert(rslt, "\"")
			table.insert(rslt, self.addressPattern)
			table.insert(rslt, "\"")
		end
		table.insert(rslt, ", ")

		if not self.location then
			table.insert(rslt, "nil")
		else
			table.insert(rslt, "\"")
			table.insert(rslt, self.location)
			table.insert(rslt, "\"")
		end
		table.insert(rslt, ", ")

		if not self.service_id then
			table.insert(rslt, "nil")
		else
			table.insert(rslt, "\"")
			table.insert(rslt, self.service_id)
			table.insert(rslt, "\"")
		end
		table.insert(rslt, ")")
		return table.concat(rslt)
	elseif self.type == "REQUEST_PROXY" then
		local rslt = {self.type, "("}
		if not self.address then
			table.insert(rslt, "nil")
		else
			table.insert(rslt, "\"")
			table.insert(rslt, self.address)
			table.insert(rslt, "\"")
		end
		table.insert(rslt, ", ")

		if not self.location then
			table.insert(rslt, "nil")
		else
			table.insert(rslt, "\"")
			table.insert(rslt, self.location)
			table.insert(rslt, "\"")
		end
		table.insert(rslt, ", ")

		if not self.service_id then
			table.insert(rslt, "nil")
		else
			table.insert(rslt, "\"")
			table.insert(rslt, self.service_id)
			table.insert(rslt, "\"")
		end
		table.insert(rslt, ")")
		return table.concat(rslt)
	elseif self.type == "DYNAMIC_SERVICE" then
		local rslt = {self.type, "("}
		if not self.location then
			table.insert(rslt, "nil")
		else
			table.insert(rslt, "\"")
			table.insert(rslt, self.location)
			table.insert(rslt, "\"")
		end
		table.insert(rslt, ", ")

		if not self.service_id then
			table.insert(rslt, "nil")
		else
			table.insert(rslt, "\"")
			table.insert(rslt, self.service_id)
			table.insert(rslt, "\"")
		end
		table.insert(rslt, ")")
		return table.concat(rslt)
	else
		return "nil"
	end
end

--action_meta.__index = action_meta

-------------
local function do_SIGNAL_EXEC(actionCtx, svcConfig, sig)
	if not actionCtx or not actionCtx.__pid then
		return false
	end
	local st, signal = pcall(require, "posix.signal")
	if st and signal then
		if type(sig) == "string" then
			sig = signal[sig]
		end

		signal.kill(actionCtx.__pid, sig)
		return true
	else
		return false
	end
end

local function do_EXEC(actionCtx,svcConfig)
	local cmd = actionCtx.cmd
	local args = actionCtx.args

	if not posix_unistd then
		_dbg.Error("CoServer", "EXEC - Unsupport EXEC('" .. cmd .. "', ...) in non-posix environment")
		return false
	end

	if not svcConfig then
		_dbg.Error("CoServer", "EXEC - Invalid service configuration passed, for: ", cmd)
		return false
	end

	local lfs = require "lfs"
	local attr = lfs.attributes(cmd)
	if not attr or not attr.permissions or not attr.permissions:match("x") then
		_dbg.Error( "CoServer", "EXEC - Invalid service command or not executable! for: ", cmd)
		return false
	end
	lfs = nil
	attr = nil

	local progArgs = {}
	local i,v
	if args then
		for i,v in pairs(args) do
			progArgs[i] = v
		end
	end
	if not progArgs[0] then
		progArgs[0] = cmd
	end

	if not svcConfig.NOT_COAF then
		if svcConfig.LOC then
			table.insert(progArgs, "--coserv:loc")
			table.insert(progArgs, svcConfig.LOC)
		end
		if svcConfig.ID then
			table.insert(progArgs, "--coserv:id")
			table.insert(progArgs, svcConfig.ID)
		end
		if svcConfig.ALT_SOCKETS then
			if type(svcConfig.ALT_SOCKETS) == "string" then
				table.insert(progArgs, "--coserv:bind")
				table.insert(progArgs, svcConfig.ALT_SOCKETS)
			elseif type(svcConfig.ALT_SOCKETS) == "table" then
				local i,v
				for i,v in ipairs(svcConfig.ALT_SOCKETS) do
					table.insert(progArgs, "--coserv:bind")
					table.insert(progArgs, v)
				end
			end
		end
	end

	local connDaemon = require "coserver.conn_daemon"
	local item
	
	if not svcConfig.NOT_COAF and svcConfig.PROXY ~= "none" and svcConfig.PROXY ~= "NONE" then
		if not svcConfig.PROXY or svcConfig.PROXY == "*" or svcConfig.PROXY == "any" or svcConfig.PROXY == "ANY" or svcConfig.PROXY == '' then
			item = connDaemon.AllowAcceptProxy( svcConfig.LOC, svcConfig.ID, nil )
		else
			item = connDaemon.AllowAcceptProxy( svcConfig.LOC, svcConfig.ID, svcConfig.PROXY )
		end
	end

	local pid = coaf.Fork(function()
		if actionCtx.type == "EXECP" then
			posix_unistd.execp(cmd, progArgs)
			os.exit(0)
		else
			posix_unistd.exec(cmd, progArgs)
			os.exit(0)
		end
	end)

	if pid and pid > 0 then
		actionCtx.__pid = pid

		local pfile
		if svcConfig.PIDFILE then
			if type(svcConfig.PIDFILE) == "string" then
				pfile = svcConfig.PIDFILE
			else
				if not svcConfig.LOC or svcConfig.LOC == "local" or svcConfig.LOC == "" then
					pfile = coaf.CoAF_pidLocation .. coaf.CoAF_pathDelim .. (svcConfig.ID or pid) .. ".pid"
				else
					pfile = coaf.CoAF_pidLocation .. coaf.CoAF_pathDelim .. svcConfig.LOC .. "." .. (svcConfig.ID or pid) .. ".pid"
				end
			end
			local f = io.open(pfile,"w")
			if not f then
				_dbg.Warn("CoServer", "EXEC - Fail to create PID file, for execute '" .. cmd .. "' and PID file is:", pfile)
				pfile = nil
			else
				f:write('' .. pid)
				f:close()
			end
		end

		local r,err = coaf.WaitPID(pid)
		
		actionCtx.__pid = nil

		if err and err:match("#.*#") then
			local st, signal = pcall(require, "posix.signal")
			if st and signal then
				signal.kill(pid, svcConfig.KILL_SIG)
			end
		end

		if item then
			connDaemon.DisallowAcceptProxy(item)
		end

		if pfile then
			os.remove(pfile)
		end

		_dbg.Info("CoServer", "EXEC - Task terminated (", cmd, ") with status:", err)
		return r
	else
		if item then
			connDaemon.DisallowAcceptProxy(item)
		end
		_dbg.Error("CoServer", "EXEC - Fail to execute '" .. cmd .. "' with returned pid:", pid)
		return false
	end
end

------------

local function do_ACCEPT_PROXY(actionCtx, svcConfig)
	if not svcConfig then
		_dbg.Error("CoServer", "ACCEPT_PROXY - Invalid service configuration")
		return false
	end
	--TODO:
	local connDaemon = require "coserver.conn_daemon"
	local loc = actionCtx.location or svcConfig.LOC or "ANY"
	local id = actionCtx.service_id or svcConfig.ID or "ANY"
	local item = connDaemon.AllowAcceptProxy( loc, id, actionCtx.addressPattern )
	if not item then
		_dbg.Error("CoServer", "ACCEPT_PROXY - Fail to accept proxy:", loc .. "/" .. id)
	else
		actionCtx.accept_item = item
	end
	return false --If use old code, this should terminate thread.
end

------------
local function do_STOP_ACCEPT_PROXY(actionCtx, svcConfig)
	if not svcConfig then
		_dbg.Error("CoServer", "STOP_ACCEPT_PROXY - Invalid service configuration")
		return false
	end
	--TODO:
	local connDaemon = require "coserver.conn_daemon"
	local loc = actionCtx.location or svcConfig.LOC or "ANY"
	local id = actionCtx.service_id or svcConfig.ID or "ANY"
	local item = actionCtx.accept_item
	if item then
		_dbg.Info("CoServer", "ACCEPT_PROXY - Disallow accept service since terminated:", loc .. "/" .. id)
		connDaemon.DisallowAcceptProxy(item)
		actionCtx.accept_item = nil
		return false --If use old code, this should terminate thread.
	else
		return false --If use old code, this should terminate thread.
	end
end

------------

local function do_REQUEST_PROXY(actionCtx, svcConfig)
	if not actionCtx.address then
		_dbg.Error("CoServer", "REQUEST_PROXY - Invalid request address")
		return false
	end

	if not svcConfig then
		_dbg.Error("CoServer", "REQUEST_PROXY - Invalid service configuration")
		return false
	end
	--TODO:
	local connDaemon = require "coserver.conn_daemon"
	local cocore = require "coserver.core"

	local loc = actionCtx.location or svcConfig.LOC or "ANY"
	local id = actionCtx.service_id or svcConfig.ID or "ANY"
	local auth_key = svcConfig.AUTHOR_KEY or cocore.configs.DEFAULT_REMOTE_KEY
	if type(auth_key) == "function" then
		auth_key = auth_key()
	end

	local conn, stat = connDaemon.RequestProxy( actionCtx.address, loc, id, svcConfig.TIMEOUT, auth_key )
	if conn then
		--NOTE here:
		stat = conn:WaitDie()
		_dbg.Info("CoServer", "REQUEST_PROXY - Requested proxy connection is disconnected:", loc .. "/" .. id)
		if stat:match("#(.*)#") then -- e.g. #CoAF-KILL#
			conn:Destroy()
			return false
		else
			return true --So try again.
		end
	else
		_dbg.Error("CoServer", "REQUEST_PROXY - Fail to request proxy:", loc .. "/" .. id, " With error:", stat )
		if stat and stat:match("#(.*)#") then -- e.g. #CoAF-KILL#
			return false
		else
			return true --So try again.
		end
	end
end

------------

local function do_DYNAMIC_SERVICE(actionCtx, svcConfig)
	if not svcConfig then
		_dbg.Error("CoServer", "DYNAMIC_SERVICE - Invalid service configuration")
		return false
	end
	local dyn_service = require "coserver.dynamic_service"

	local loc = actionCtx.location or svcConfig.LOC or "ANY"
	local id = actionCtx.service_id or svcConfig.ID or "ANY"

	dyn_service.StartService(loc,id)
end

local function do_STOP_DYNAMIC_SERVICE(actionCtx, svcConfig)
	if not svcConfig then
		_dbg.Error("CoServer", "DYNAMIC_SERVICE - Invalid service configuration")
		return false
	end
	local dyn_service = require "coserver.dynamic_service"

	local loc = actionCtx.location or svcConfig.LOC or "ANY"
	local id = actionCtx.service_id or svcConfig.ID or "ANY"

	dyn_service.StopService(loc,id)
end

------------

function _M.EXEC(cmd,args)
	local actionCtx = {
		IS_COSERVER_ACTION = true,
		type = "EXEC",
		cmd = cmd,
		args = args,
		doAction = do_EXEC,
		doSignalAction = do_SIGNAL_EXEC
	}
	setmetatable(actionCtx, action_meta)
	return actionCtx
end

------------

function _M.EXECP(cmd,args)
	local actionCtx = {
		IS_COSERVER_ACTION = true,
		type = "EXECP",
		cmd = cmd,
		args = args,
		doAction = do_EXEC,
		doSignalAction = do_SIGNAL_EXEC
	}
	setmetatable(actionCtx, action_meta)
	return actionCtx
end

------------

function _M.ACCEPT_PROXY(addressPattern, svcLoc, svcId)
	local actionCtx = {
		IS_COSERVER_ACTION = true,
		IS_ONESHOT = true,
		type = "ACCEPT_PROXY",
		addressPattern = addressPattern,
		location = svcLoc,
		service_id = svcId,
		doAction = do_ACCEPT_PROXY,
		doStopAction = do_STOP_ACCEPT_PROXY
	}
	setmetatable(actionCtx, action_meta)
	return actionCtx
end

------------

function _M.REQUEST_PROXY(address, svcLoc, svcId)
	local actionCtx = {
		IS_COSERVER_ACTION = true,
		type = "REQUEST_PROXY",
		address = address,
		location = svcLoc,
		service_id = svcId,
		doAction = do_REQUEST_PROXY
	}
	setmetatable(actionCtx, action_meta)
	return actionCtx
end

------------
function _M.DYNAMIC_SERVICE(svcLoc, svcId)
	local actionCtx = {
		IS_COSERVER_ACTION = true,
		IS_ONESHOT = true, --If oneshot, do not start thread.
		type = "DYNAMIC_SERVICE",
		location = svcLoc,
		service_id = svcId,
		doAction = do_DYNAMIC_SERVICE,
		doStopAction = do_STOP_DYNAMIC_SERVICE
	}
	setmetatable(actionCtx, action_meta)
	return actionCtx
end

------------

return _M
