--[==[

The Co-Socket for CoAF (Co-Application Framework)

The cosocket module is similar to lua-socket, but, it works
asynchronously. If you are calling cosocket APIs in a 
coroutine context and called API may be blocked, it 
should yield current coroutine, util unblocked or
timeouted, it returns.

The cosocket module depends 'lua-ev' module.

Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.

]==]

local _M = {
	default_close_on_exec = nil, --TODO: Now global set this flag to mark socket 'close-on-exec'
}

--Original lua socket.
local socket = require "socket"

--Require lua-ev module.
local ev = require "ev"

local tcp_m = require "cosocket.tcp"		--get TCP patch module.
local udp_m = require "cosocket.udp"		--get UDP patch module.
local unix_sock_f = require "cosocket.unix"		--get UNIX patch module.

-- No overrided functions/properties, use default.
function _M.__index(m, key)
	return socket[key]
end

-----------------------------------------------------------------------------

function _M.flag_close_on_exec(fd, en)
	if nil == en then
		return --if nil, not to set anything.
	end

	local compa = require "coaf.compatible"
	local posix_fcntl = compa.CheckPosix( "posix.fcntl" )
	if not posix_fcntl then
		return false
	end
	local FD_CLOEXEC = posix_fcntl.FD_CLOEXEC or 1
	local flag = posix_fcntl.fcntl( fd, posix_fcntl.F_GETFD )
	if compa.lua_version >= 5.3 then
		if not en then
			flag = flag & (~FD_CLOEXEC)
		else
			flag = flag | FD_CLOEXEC
		end
	elseif compa.bit then
		if not en then
			flag = compa.bit.band( flag, _M.bit.bnot(FD_CLOEXEC) )
		else
			flag = compa.bit.bor( flag, FD_CLOEXEC )
		end
	end
	posix_fcntl.fcntl( fd, posix_fcntl.F_SETFD, flag )
	return true
end

-----------------------------------------------------------------------------
-- Fixed async methods.
-----------------------------------------------------------------------------

function _M.tcp()
	local tcp, err = socket.tcp()
	if not tcp then
		return tcp,err
	end
	return tcp_m.build_socket(tcp)
end

-- Added a new function to get unix domain socket.
function _M.unix()
	return unix_sock_f()
end

function _M.tcp6()
	local tcp, err = socket.tcp6()
	if not tcp then
		return tcp,err
	end
	return tcp_m.build_socket(tcp)
end

function _M.udp()
	local udp, err = socket.udp()
	if not udp then
		return udp,err
	end
	return udp_m.build_socket(udp)
end

function _M.udp6()
	local udp, err = socket.udp6()
	if not udp then
		return udp,err
	end
	return udp_m.build_socket(udp)
end

--TODO: Replace it with a non-blocked implementation.
_M.dns = socket.dns

function _M.select(recvt, sendt, timeout )
--TODO: implement it to async.
	return socket.select(recvt, sendt, timeout)
end

function _M.sleep(time)
	local coro, ismain = coroutine.running()
	--Note: sleep(0) is important for task switch.
	if not time or time < 0 then
		return
	else
		local task = ev.Timer.new( function(loop,tmr,rev)
			tmr:stop(loop)
			if ismain then
				--print( "!!! Detail of socket.sleep: in main task and unloop")
				loop:unloop()
			else
				local r, st = coroutine.resume(coro)
				if not r then
					print("*** Thread error:", st )
				end
			end
		end, time, 0 )
		if task then
			local loop = _G.CoAF_LOOP or ev.Loop.default
			task:start(loop)
			if ismain then
				loop:loop()
			else
				local r,flag = coroutine.yield(coro)
				if not r and flag == "#CoAF-KILL#" then
					if task:is_pending() then task:clear_pending(loop) end
					task:stop(loop)
				end
			end
		end
	end
end

-----------------------------------------------------------------------------
-- Exported auxiliar functions (Copied from lua-socket and changed)
-----------------------------------------------------------------------------
function _M.connect(address, port, laddress, lport, family)
	
	--TODO: How to process laddress and lport ?

	local sock,err
	if family == "inet6" then
		sock,err = _M.tcp6()
	else
		sock,err = _M.tcp()
	end
	if not sock then return nil, err end
	local x
	x,err = sock:connect(address,port)
	if not x then
		sock:close()
		return nil, err
	else
		return sock
	end
end

function _M.connect4(address, port, laddress, lport)
    return _M.connect(address, port, laddress, lport, "inet")
end

function _M.connect6(address, port, laddress, lport)
    return _M.connect(address, port, laddress, lport, "inet6")
end

function _M.bind(host, port, backlog)
	if host == "*" then host = "0.0.0.0" end
	
	local addrinfo, err = socket.dns.getaddrinfo(host);
	
    if not addrinfo then return nil, err end
    local sock, res
    err = "no info on address"
    for i, alt in base.ipairs(addrinfo) do
        if alt.family == "inet" then
            sock, err = _M.tcp()
        else
            sock, err = _M.tcp6()
        end
        if not sock then return nil, err end
        sock:setoption("reuseaddr", true)
        res, err = sock:bind(alt.addr, port)
        if not res then 
            sock:close()
        else 
            res, err = sock:listen(backlog)
            if not res then 
                sock:close()
            else
                return sock
            end
        end 
    end
    return nil, err
end

return setmetatable(_M, _M)
