--[==[

The UDP Co-Socket for CoAF (Co-Application Framework)

The cosocket module is similar to lua-socket, but, it works
asynchronously. If you are calling cosocket APIs in a 
coroutine context and called API may be blocked, it 
should yield current coroutine, util unblocked or
timeouted, it returns.

The cosocket module depends 'lua-ev' module.

Copyright (C) 2018, KILOVIEW Electronics, Co., Ltd.

]==]

local ev = require "ev"
local socket = require "socket"

local _M = {}

local udp_meta = {}

local function TRACE_CORO_RESUME(co,...)
	local r, st = coroutine.resume(co, ...)
	if not r then
		print("*** Thread error:", st)
	end
end

function udp_meta.getfd(udp)
	return udp.sock:getfd()
end

function udp_meta.setfd(udp, fd)
	--TODO: close-on-exec flag.
	return udp.sock:setfd(fd)
end

function udp_meta.dirty(udp)
	return udp.sock:dirty()
end

function udp_meta.cleanup(udp, mode)
	local clear_tx, clear_rx
	if not mode or mode == "both" then
		clear_tx = true
		clear_rx = true
	elseif mode == "send" then
		clear_tx = true
		clear_rx = false
	elseif mode == "receive" then
		clear_tx = false
		clear_rx = true
	end

	if clear_tx and udp.send_task then
		udp.send_task.cancel()
	end

	if clear_rx and udp.recv_task then
		udp.recv_task.cancel()
	end
end

function udp_meta.close(udp)
	udp:cleanup("both")
	udp.sock:close()
end

function udp_meta.getpeername(udp)
	return udp.sock:getpeername()
end

function udp_meta.setpeername(udp,addr,port)
	return udp.sock:setpeername(addr,port)
end

function udp_meta.getsockname(udp)
	return udp.sock:getsockname()
end

function udp_meta.setsockname(udp,addr,port)
	return udp.sock:setsockname(addr,port)
end

--[[
function udp_meta.getstats(udp)
	return udp.sock:getstats()
end
]]

function udp_meta.receivefrom(udp, size)
	if udp.recv_task then
		return nil, "busy"
	end

	local coro, ismain = coroutine.running()

	udp.sock:settimeout(0)
	if udp.timeout == 0 then
		return udp.sock:receivefrom(size)
	end

	local mainResult = { nil, "timeout" } --TODO: Why can't set correct upvalue?

	-- Wait read.
	local task = ev.IO.new( function(loop, wa)
		wa:stop(loop)
		--Clear timeout timer.
		if udp.recv_task and udp.recv_task.tmo_task then
			if udp.recv_task.tmo_task:is_pending() then
				udp.recv_task.tmo_task:clear_pending(loop)
			end
			udp.recv_task.tmo_task:stop(loop)
		end

		udp.recv_task = nil

		local data, err_or_ip, port = udp.sock:receivefrom(size)
		if ismain then
			mainResult = {data, err_or_ip, port}
			--print("**** recvfrom()/1: unloop")
			loop:unloop()
		else
			TRACE_CORO_RESUME(coro, data, err_or_ip, port)
		end
	end, udp.sock:getfd(), ev.READ)

	if task then
		local tmoTask
		-- That, if timeout < 0 or nil, wait infinity.
		if udp.timeout and udp.timeout > 0 then
			tmoTask = ev.Timer.new( function(loop,tmr)
				tmr:stop(loop)

				if task:is_pending() then
					task:clear_pending(loop)
				end
				task:stop(loop)
				udp.recv_task = nil
				if ismain then
					mainResult = {nil,"timeout"} --TODO: Why can't set correct upvalue?
					--print("**** recvfrom()/2: unloop")
					loop:unloop()
				else
					TRACE_CORO_RESUME(coro, nil, "timeout")
				end
			end, udp.timeout, 0)
		end

		local loop = _G.CoAF_LOOP or ev.Loop.default
		udp.recv_task = { 
			io_task = task,
			tmo_task = tmoTask,
			-- Cancel (for closing)
			cancel = function()
				if task:is_pending() then
					task:clear_pending(loop)
				end
				task:stop(loop)

				if udp.recv_task.tmo_task then
					if udp.recv_task.tmo_task:is_pending() then
						udp.recv_task.tmo_task:clear_pending(loop)
					end
					udp.recv_task.tmo_task:stop(loop)
				end
				udp.recv_task = nil
				if ismain then
					mainResult = {nil,"closed"} --TODO: Why can't set correct upvalue?
					--print("**** recvfrom()/3: unloop")
					loop:unloop()
				else
					TRACE_CORO_RESUME(coro, nil, "closed")
				end
			end
		}

		task:start(loop)
		if tmoTask then
			tmoTask:start(loop)
		end
		if ismain then
			loop:loop()
			return table.unpack(mainResult)
		else
			local r,err,port = coroutine.yield(coro)
			if not r and err == "#CoAF-KILL#" then
				if task:is_pending() then task:clear_pending(loop) end
				task:stop(loop)
				if tmoTask then
					if tmoTask:is_pending() then tmoTask:clear_pending(loop) end
					tmoTask:stop(loop)
				end
				tcp.recv_task = nil
			end
			return r,err,port
		end
	else
		return nil, "fail to start async recv"
	end
end

function udp_meta.receive(udp, size)
	local r1, r2, r3 = udp_meta.receivefrom(udp, size)
	if r1 then
		return r1
	else
		return r1, r2
	end
end


function udp_meta.send(udp, data)
	--Since udp never block, directly send.
	return udp.sock:send(data)
end

function udp_meta.sendto(udp, data, ip, port)
	--Since udp never block, directly send.
	return udp.sock:sendto(data, ip, port)
end

function udp_meta.setoption(udp,option,val)
	return udp.sock:setoption(option,val)
end

--[[
function udp_meta.setstats(udp,received,sent,age)
	return udp.sock:setstats(received,sent,age)
end
]]

function udp_meta.settimeout(udp,value,mode)
	udp.timeout = value or udp.timeout
end

function _M.build_socket(udp)
	local socket_core = require "cosocket.core"
	udp:settimeout(0)	--non-block socket.
	socket_core.flag_close_on_exec( udp:getfd(), socket_core.default_close_on_exec )
	local udp_sock = {
		sock = udp,
		timeout = 30,	-- The default action timeout!
		recv_task = nil,
		send_task = nil
	}
	return setmetatable(udp_sock, {__index=udp_meta})
end

return _M
